from logging import Logger


class Configurations(object):
    def __init__(
        self,
        required_services=None,
        service_name=None,
        persistence_name=None,
        store_id=None,
        logger=None,
        base_url=None,
        api_key=None
    ):
        # type: (str, str, str, int, Logger, str, str) -> None

        self.required_services = required_services
        self.service_name = service_name
        self.persistence_name = persistence_name
        self.store_id = store_id
        self.logger = logger
        self.base_url = base_url
        self.api_key = api_key
        self.endpoints = Endpoints()


class Endpoints(object):
    def __init__(self):
        self.retrieve_loyalty_id = None
        self.get_benefit = None
        self.burn_benefit = None
        self.unlock_benefit = None



from ._Action import Action
from ._Conditions import Conditions
from ._Exclusions import Exclusions
from ._Customer import Customer
from ._Benefit import Benefit
from ._AdditionalInfo import AdditionalInfo
from ._BenefitInfo import BenefitInfo
from ._CustomerInfo import CustomerInfo



from production.box import ShowOnKitchenFilterBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class ShowOnKitchenFilterBoxBuilder(BoxBuilder):
    def __init__(self, product_repository, loggers):
        super(ShowOnKitchenFilterBoxBuilder, self).__init__(loggers)
        self.product_repository = product_repository

    def build(self, config):
        # type: (BoxConfiguration) -> ShowOnKitchenFilterBox

        inherit_comments = self.get_extra(config, "InheritComments", "false").lower() == "true"
        return ShowOnKitchenFilterBox(
            config.name,
            config.sons,
            self.product_repository,
            inherit_comments,
            self.get_logger(config),
        )



import sysactions


def get_store_id():
    # type: (...) -> str

    return sysactions.get_storewide_config("Store.Id")


def get_boh_url():
    # type: (...) -> str

    return sysactions.get_storewide_config("BackOffice.ServerURL")


def get_boh_api_key():
    # type: (...) -> str

    return sysactions.get_storewide_config("BackOffice.ApiKey")



from production.box import RecalledOrderStampBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class RecalledOrderStampBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(RecalledOrderStampBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> RecalledOrderStampBoxBuilder
        filter_tag = config.extra_config["FilterTag"] if "FilterTag" in config.extra_config else "done"
        ignore_prod_state_config = config.extra_config.get("IgnoreProdState")
        ignore_prod_state = ignore_prod_state_config.lower() == "true" if ignore_prod_state_config else False

        return RecalledOrderStampBox(
            config.name,
            config.sons,
            filter_tag,
            ignore_prod_state,
            self.get_logger(config)
        )



from ..http import parse_authorization_header
from ..http import parse_www_authenticate_header
from ..utils import cached_property


class AuthorizationMixin(object):
    """Adds an :attr:`authorization` property that represents the parsed
    value of the `Authorization` header as
    :class:`~werkzeug.datastructures.Authorization` object.
    """

    @cached_property
    def authorization(self):
        """The `Authorization` object in parsed form."""
        header = self.environ.get("HTTP_AUTHORIZATION")
        return parse_authorization_header(header)


class WWWAuthenticateMixin(object):
    """Adds a :attr:`www_authenticate` property to a response object."""

    @property
    def www_authenticate(self):
        """The `WWW-Authenticate` header in a parsed form."""

        def on_update(www_auth):
            if not www_auth and "www-authenticate" in self.headers:
                del self.headers["www-authenticate"]
            elif www_auth:
                self.headers["WWW-Authenticate"] = www_auth.to_header()

        header = self.headers.get("www-authenticate")
        return parse_www_authenticate_header(header, on_update)



import datetime

from commons.dto import ReportColumnDefinition, AlignTypes
from commons.report import Formatter
from commons.util import TableReport
from reports_app.loginsummary.dto import LoginSummaryDto  # noqa
from reports_app.loginsummary.dto import LoginSummaryHeaderDto  # noqa


class LoginFormatter(Formatter):
    def __init__(self):
        self.report_lines = []
        self.translated_messages = []

    def format_report(self, login_summary_dto, translator):
        # type: (LoginSummaryDto) -> None
        self.translated_messages = translator.translate_labels(["$LOGIN",
                                                                "$STORE",
                                                                "$POS",
                                                                "$OPERATOR",
                                                                "$CURRENT_DATE",
                                                                "$BUSINESS_DATE",
                                                                "$LOGIN_TIME",
                                                                "$POS_TYPE",
                                                                "$AUTHORIZER",
                                                                "$INITIAL_VALUE"
                                                                ],
                                                               login_summary_dto.login_summary_header_dto.pos_id)

        report = self._append_header(login_summary_dto.login_summary_header_dto)
        report += self._append_body(login_summary_dto.login_summary_body_dto)

        report_bytes = report.decode("utf-8").encode("utf-8")
        return report_bytes

    def _append_header(self, login_summary_header_dto):
        # type: (LoginSummaryHeaderDto) -> str
        self._append_line_separator()
        self._append_title()
        self._append_store_pos(login_summary_header_dto.store_id, login_summary_header_dto.pos_id)
        self._append_operator(login_summary_header_dto.operator_id)
        self._append_current_date()
        self._append_business_date(login_summary_header_dto.business_date)

        table_formatter = TableReport([
            ReportColumnDefinition(width=17, fill_with_char='.'),
            ReportColumnDefinition(width=21, before_text=': ')
        ])

        return table_formatter.format(self.report_lines)

    def _append_body(self, login_summary_body_dto):
        self.report_lines = []

        self._append_empty_line()

        self._append_pos_type(login_summary_body_dto.pos_type)
        self._append_authorizer(login_summary_body_dto.authorizer)

        self._append_initial_value(login_summary_body_dto.initial_value)
        self._append_line_separator()

        table_formatter = TableReport([
            ReportColumnDefinition(width=17, fill_with_char='.'),
            ReportColumnDefinition(width=21, before_text=': ')
        ])

        return table_formatter.format(self.report_lines)

    def _append_line_separator(self):
        self.report_lines.append([ReportColumnDefinition(
            fill_with_char='='
        )])

    def _append_title(self):
        title = self.translated_messages["$LOGIN"]

        self.report_lines.append([ReportColumnDefinition(
            text=title,
            align=AlignTypes.CENTER
        )])

    def _append_store_pos(self, store_id, pos_id):
        self.report_lines.append([self.translated_messages["$STORE"] + "/" + self.translated_messages["$POS"],
                                  (str(store_id) + '/' + str(pos_id))])

    def _append_operator(self, operator_id):
        self.report_lines.append([self.translated_messages["$OPERATOR"], str(operator_id)])

    def _append_current_date(self):
        self.report_lines.append([self.translated_messages["$CURRENT_DATE"],
                                  datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")])

    def _append_business_date(self, business_date):
        self.report_lines.append([self.translated_messages["$BUSINESS_DATE"], business_date])

    def _append_empty_line(self):
        self.report_lines.append([ReportColumnDefinition(fill_with_char=' ')])

    def _append_authorizer(self, authorizer):
        self.report_lines.append([self.translated_messages["$AUTHORIZER"], str(authorizer)])

    def _append_pos_type(self, pos_type):
        self.report_lines.append([self.translated_messages["$POS_TYPE"], str(pos_type)])

    def _append_initial_value(self, initial_value):
        self.report_lines.append([self.translated_messages["$INITIAL_VALUE"], ("R$ " + str(initial_value))])



from ._ValidationUtil import (
    check_if_business_day_is_opened,
    check_if_operator_is_logged,
    check_if_order_is_in_progress,
    check_if_order_is_totaled,
    check_if_order_is_in_progress_or_totaled,
    check_if_pos_id_is_valid,
    check_if_payment_type_is_valid,
    check_if_payment_amount_is_valid,
    check_if_due_amount_is_zero,
    check_business_period,
    validate_api_key,
    get_pos_id_from_pos_mapper,
    get_abstracted_pos_id_from_pos_mapper,
    validate_pos_id,
    check_has_due_amount,
    check_if_request_order_id_is_equal_current_order_id,
)




# -*- coding: utf-8 -*-

import json

import sysactions
from application.model.benefit import BenefitList
from application.model.exceptions import OrderNotFoundException, OrderPictureException

from application.model.order import Order
from typing import (
    Dict,
    List,
    Optional,
    Union,
)
from xml.etree import cElementTree as eTree
from xml.etree.ElementTree import Element

from messagebus import MessageBus, Message, DataType, DefaultToken
from msgbus import TK_SLCTRL_OMGR_ORDERPICT
from posot import OrderTaker, OrderTakerException
from mwhelper import ModelService

from application.model import Configurations, NDiscountModelJsonEncoder
from application.repository import NDiscountRepository


class OrderService(object):
    def __init__(self, configs, repository, message_bus):
        # type: (Configurations, NDiscountRepository, MessageBus) -> None

        self.configs = configs
        self.logger = configs.logger
        self.repository = repository
        self.message_bus = message_bus

    @staticmethod
    def get_burnt_items(applied_benefits):
        # type: (eTree) -> Dict

        burnt_items = {}
        for benefit in applied_benefits:
            current_burnt_items = benefit["burnt_items"]
            for key in current_burnt_items.keys():
                if key in burnt_items:
                    burnt_items[key] += current_burnt_items[key]
                else:
                    burnt_items[key] = current_burnt_items[key]

        return burnt_items

    def get_applied_benefits_from_order_picture(self, pos_id, order_picture=None):
        # type: (int, Optional[eTree]) -> List

        benefit_list = self.get_benefit_list_from_order_picture(pos_id, order_picture)
        if not benefit_list:
            return []

        return json.loads(benefit_list.get("benefitList", "[]"))

    def get_benefit_list_from_order_picture(self, pos_id, order_picture=None):
        # type: (int, Optional[eTree]) -> Dict

        try:
            if order_picture is None:
                order_picture = self.get_order_picture(pos_id)

            value = order_picture.find(".//OrderProperty[@key=\"BENEFIT_LIST\"]").get("value")
            return json.loads(value)
        except OrderTakerException:
            pass
        except AttributeError:
            pass
        return {}

    @staticmethod
    def save_benefit_list_custom_property(pos_ot, benefit_list):
        # type: (OrderTaker, BenefitList) -> None

        benefit_list_json = json.dumps(benefit_list, encoding="utf-8", cls=NDiscountModelJsonEncoder)
        pos_ot.setOrderCustomProperty("BENEFIT_LIST", benefit_list_json)

    def get_order_picture(self, pos_id, order_id=None):
        # type: (Union[str, int], Optional[str]) -> Optional[Element]

        data = "{}\0{}".format(str(pos_id), str(order_id or ''))
        message = Message(TK_SLCTRL_OMGR_ORDERPICT, DataType.param, data, 30000000)
        reply = self.message_bus.send_message("ORDERMGR{}".format(str(pos_id)), message)

        if reply.token == DefaultToken.TK_SYS_ACK:
            try:
                split_reply = reply.data.split("\0")
                if not split_reply[2]:
                    return None
                order_picture = eTree.XML(split_reply[2])
                if order_id:
                    order_picture = order_picture.findall('.//Order')[0]
                return order_picture
            except AttributeError:
                return None

        raise OrderPictureException("Error getting OrderPicture: {} - {}".format(reply.token, reply.data))

    def get_order(self, pos_id, model=None, create_order=False, order_picture=None, table_id=None):
        # type: (int, Element, bool, Element, str) -> Order

        if not order_picture:
            order_picture = self.get_order_picture(pos_id)

        state = order_picture.get("state", "") if order_picture else ""
        if state not in ["IN_PROGRESS", "TOTALED"]:
            if not create_order:
                raise OrderNotFoundException()

            order_picture = self._create_new_order(pos_id, table_id)

        sale_lines = order_picture.findall(".//SaleLine")

        return Order(order_picture, sale_lines, model)

    @staticmethod
    def void_order(pos_id, pos_ot):
        # type: (int, OrderTaker) -> None

        pos_ot.voidOrder(pos_id)

    @staticmethod
    def void_line(pos_id, pos_ot, lines_to_void):
        # type: (int, OrderTaker, List) -> None

        if len(lines_to_void) > 0:
            lines_to_void = "|".join([str(x) for x in lines_to_void])
            pos_ot.voidLine(pos_id, lines_to_void)

    @staticmethod
    def clear_discounts(pos_ot, used_discount_ids):
        # type: (OrderTaker, List) -> None

        for discount_id in used_discount_ids:
            pos_ot.clearDiscount(discount_id)

    @staticmethod
    def clear_tenders(pos_id, pos_ot, used_order_tender_ids):
        # type: (int, OrderTaker, List) -> None

        for order_tender_id in used_order_tender_ids:
            pos_ot.clearTenders(pos_id, None, order_tender_id)

    def _create_new_order(self, pos_id, table_id=None):
        # type: (int, str) -> Element

        # TODO: Review parameters when there is no Order (PriceList and SaleType)
        model = sysactions.get_model(pos_id)
        pos_ot = sysactions.get_posot(model)
        pos_ot.blkopnotify = True
        sale_type = "EAT_IN"
        model_service = ModelService(pos_id=pos_id)
        order = pos_ot.createOrder(
            posid=pos_id,
            pricelist=model_service.get_price_list_from_model(
                sale_type=sale_type,
            ),
            saletype=sale_type,
        )

        if table_id:
            pos_ot.setOrderCustomProperty("TABLE_ID", value=table_id, orderid=order.get("orderId"))

        pos_ot.blkopnotify = False

        return self.get_order_picture(pos_id)



from abc import ABCMeta, abstractmethod


class ProductionCommand(object):
    __metaclass__ = ABCMeta

    def __init__(self, order_id, view):
        self.order_id = order_id
        self.view = view

    @abstractmethod
    def get_hash_extra_value(self):
        # type: () -> str
        raise NotImplementedError()

    def __hash__(self):
        h = 0xcbf29ce484222325
        for c in "{}{}{}".format(self.order_id, self.view, self.get_hash_extra_value()):
            h *= 1099511628211
            h ^= ord(c)
        return h



import os
import sys

import sysactions
from configurator import Configurator
from flaskmessagehandler import FlaskMessageHandler
from helper import import_pydevd
from httpmessagehandler import HttpMessageHandler
from mbcontextmessagehandler import MbContextMessageBus
from messagehandlerbuilder import ServiceMessageHandlerBuilder
from model.config import Config
from msgbus import MBEasyContext, MBException
from syserrors import SE_MB_ERR, SE_SUCCESS


def main():
    # type: () -> FlaskMessageHandler

    loader_path = os.environ["LOADERCFG"]

    import_pydevd(loader_path, 9178)

    configurator = Configurator(loader_path)
    configs = configurator.read_loader_configs()
    logger = configurator.configure_logger()

    mb_context = MBEasyContext(configs.component_name)
    message_bus = MbContextMessageBus(mb_context)
    sysactions.mbcontext = mb_context

    _wait_for_message_bus_required_services(message_bus, configs)

    service_message_handler_builder = ServiceMessageHandlerBuilder(logger, configs, message_bus, flask_server=True)
    http_message_handler = HttpMessageHandler(service_message_handler_builder, configs.cors_origin_hosts)
    flask_message_handler = FlaskMessageHandler(http_message_handler)

    logger.debug("Starting flask [{}] ...".format(configs.component_name))
    return flask_message_handler.run(configs.component_name, configs.host, configs.port)


def _wait_for_message_bus_required_services(message_bus, configs):
    # type: (MbContextMessageBus, Config) -> None

    try:
        exported_services = "{}:{}".format(configs.component_name, configs.component_name)
        status = message_bus.mbcontext.MB_EasyWaitStart(exported_services, configs.required_services)
        if status != SE_SUCCESS:
            sys.exit(status)
    except MBException:
        sys.exit(SE_MB_ERR)


if __name__ == "__main__":
    main()



from typing import List  # noqa
from _CashReportBodyLineDto import CashReportBodyLineDto  # noqa


class CashReportBodyDto(object):
    def __init__(self, cash_report_body_lines_dto, report_type):
        # type:(List[CashReportBodyLineDto], unicode) -> None
        self.cash_report_body_lines_dto = cash_report_body_lines_dto  # type:(List[CashReportBodyLineDto])
        self.report_type = report_type



#-----------------------------------------------------------------
# plyparser.py
#
# PLYParser class and other utilites for simplifying programming
# parsers with PLY
#
# Eli Bendersky [https://eli.thegreenplace.net/]
# License: BSD
#-----------------------------------------------------------------

import warnings

class Coord(object):
    """ Coordinates of a syntactic element. Consists of:
            - File name
            - Line number
            - (optional) column number, for the Lexer
    """
    __slots__ = ('file', 'line', 'column', '__weakref__')
    def __init__(self, file, line, column=None):
        self.file = file
        self.line = line
        self.column = column

    def __str__(self):
        str = "%s:%s" % (self.file, self.line)
        if self.column: str += ":%s" % self.column
        return str


class ParseError(Exception): pass


class PLYParser(object):
    def _create_opt_rule(self, rulename):
        """ Given a rule name, creates an optional ply.yacc rule
            for it. The name of the optional rule is
            <rulename>_opt
        """
        optname = rulename + '_opt'

        def optrule(self, p):
            p[0] = p[1]

        optrule.__doc__ = '%s : empty\n| %s' % (optname, rulename)
        optrule.__name__ = 'p_%s' % optname
        setattr(self.__class__, optrule.__name__, optrule)

    def _coord(self, lineno, column=None):
        return Coord(
                file=self.clex.filename,
                line=lineno,
                column=column)

    def _token_coord(self, p, token_idx):
        """ Returns the coordinates for the YaccProduction objet 'p' indexed
            with 'token_idx'. The coordinate includes the 'lineno' and
            'column'. Both follow the lex semantic, starting from 1.
        """
        last_cr = p.lexer.lexer.lexdata.rfind('\n', 0, p.lexpos(token_idx))
        if last_cr < 0:
            last_cr = -1
        column = (p.lexpos(token_idx) - (last_cr))
        return self._coord(p.lineno(token_idx), column)

    def _parse_error(self, msg, coord):
        raise ParseError("%s: %s" % (coord, msg))


def parameterized(*params):
    """ Decorator to create parameterized rules.

    Parameterized rule methods must be named starting with 'p_' and contain
    'xxx', and their docstrings may contain 'xxx' and 'yyy'. These will be
    replaced by the given parameter tuples. For example, ``p_xxx_rule()`` with
    docstring 'xxx_rule  : yyy' when decorated with
    ``@parameterized(('id', 'ID'))`` produces ``p_id_rule()`` with the docstring
    'id_rule  : ID'. Using multiple tuples produces multiple rules.
    """
    def decorate(rule_func):
        rule_func._params = params
        return rule_func
    return decorate


def template(cls):
    """ Class decorator to generate rules from parameterized rule templates.

    See `parameterized` for more information on parameterized rules.
    """
    issued_nodoc_warning = False
    for attr_name in dir(cls):
        if attr_name.startswith('p_'):
            method = getattr(cls, attr_name)
            if hasattr(method, '_params'):
                # Remove the template method
                delattr(cls, attr_name)
                # Create parameterized rules from this method; only run this if
                # the method has a docstring. This is to address an issue when
                # pycparser's users are installed in -OO mode which strips
                # docstrings away.
                # See: https://github.com/eliben/pycparser/pull/198/ and
                #      https://github.com/eliben/pycparser/issues/197
                # for discussion.
                if method.__doc__ is not None:
                    _create_param_rules(cls, method)
                elif not issued_nodoc_warning:
                    warnings.warn(
                        'parsing methods must have __doc__ for pycparser to work properly',
                        RuntimeWarning,
                        stacklevel=2)
                    issued_nodoc_warning = True
    return cls


def _create_param_rules(cls, func):
    """ Create ply.yacc rules based on a parameterized rule function

    Generates new methods (one per each pair of parameters) based on the
    template rule function `func`, and attaches them to `cls`. The rule
    function's parameters must be accessible via its `_params` attribute.
    """
    for xxx, yyy in func._params:
        # Use the template method's body for each new method
        def param_rule(self, p):
            func(self, p)

        # Substitute in the params for the grammar rule and function name
        param_rule.__doc__ = func.__doc__.replace('xxx', xxx).replace('yyy', yyy)
        param_rule.__name__ = func.__name__.replace('xxx', xxx)

        # Attach the new method to the class
        setattr(cls, param_rule.__name__, param_rule)



from functions_authorization import get_function_authorization_level_from_config
from get_keyboard_comments_config import get_keyboard_comments_config
from get_min_value_to_ask_bordereau_justify_config import get_min_value_to_ask_bordereau_justify_config
from show_numpad_for_customer_name import show_numpad_for_customer_name



from abc import ABCMeta, abstractmethod
from typing import Optional, Dict

from ._HttpClient import HttpClient


class DefaultHttpClient(HttpClient):
    __metaclass__ = ABCMeta

    def __init__(self, http_client, base_url, default_headers=None):
        # type: (HttpClient, str, Optional[Dict[str, str]]) -> None
        super(DefaultHttpClient, self).__init__()
        self.http_client = http_client
        self.base_url = base_url
        self.default_headers = default_headers

    @abstractmethod
    def send(self, request):
        request.url = self.base_url + request.url
        if self.default_headers is not None:
            for header in self.default_headers.keys():
                if header not in request.headers:
                    request.headers[header] = self.default_headers[header]
        self.http_client.send(request)



# Copyright 2012-2019, Andrey Kislyuk and argcomplete contributors.
# Licensed under the Apache License. See https://github.com/kislyuk/argcomplete for more info.

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import subprocess
from .compat import str, sys_encoding

def _call(*args, **kwargs):
    try:
        return subprocess.check_output(*args, **kwargs).decode(sys_encoding).splitlines()
    except subprocess.CalledProcessError:
        return []

class ChoicesCompleter(object):
    def __init__(self, choices):
        self.choices = choices

    def _convert(self, choice):
        if isinstance(choice, bytes):
            choice = choice.decode(sys_encoding)
        if not isinstance(choice, str):
            choice = str(choice)
        return choice

    def __call__(self, **kwargs):
        return (self._convert(c) for c in self.choices)

EnvironCompleter = ChoicesCompleter(os.environ)

class FilesCompleter(object):
    """
    File completer class, optionally takes a list of allowed extensions
    """

    def __init__(self, allowednames=(), directories=True):
        # Fix if someone passes in a string instead of a list
        if isinstance(allowednames, (str, bytes)):
            allowednames = [allowednames]

        self.allowednames = [x.lstrip("*").lstrip(".") for x in allowednames]
        self.directories = directories

    def __call__(self, prefix, **kwargs):
        completion = []
        if self.allowednames:
            if self.directories:
                files = _call(["bash", "-c", "compgen -A directory -- '{p}'".format(p=prefix)])
                completion += [f + "/" for f in files]
            for x in self.allowednames:
                completion += _call(["bash", "-c", "compgen -A file -X '!*.{0}' -- '{p}'".format(x, p=prefix)])
        else:
            completion += _call(["bash", "-c", "compgen -A file -- '{p}'".format(p=prefix)])
            anticomp = _call(["bash", "-c", "compgen -A directory -- '{p}'".format(p=prefix)])
            completion = list(set(completion) - set(anticomp))

            if self.directories:
                completion += [f + "/" for f in anticomp]
        return completion

class _FilteredFilesCompleter(object):
    def __init__(self, predicate):
        """
        Create the completer

        A predicate accepts as its only argument a candidate path and either
        accepts it or rejects it.
        """
        assert predicate, "Expected a callable predicate"
        self.predicate = predicate

    def __call__(self, prefix, **kwargs):
        """
        Provide completions on prefix
        """
        target_dir = os.path.dirname(prefix)
        try:
            names = os.listdir(target_dir or ".")
        except:
            return  # empty iterator
        incomplete_part = os.path.basename(prefix)
        # Iterate on target_dir entries and filter on given predicate
        for name in names:
            if not name.startswith(incomplete_part):
                continue
            candidate = os.path.join(target_dir, name)
            if not self.predicate(candidate):
                continue
            yield candidate + "/" if os.path.isdir(candidate) else candidate

class DirectoriesCompleter(_FilteredFilesCompleter):
    def __init__(self):
        _FilteredFilesCompleter.__init__(self, predicate=os.path.isdir)

class SuppressCompleter(object):
    """
    A completer used to suppress the completion of specific arguments
    """

    def __init__(self):
        pass

    def suppress(self):
        """
        Decide if the completion should be suppressed
        """
        return True



# -*- coding: utf-8 -*-
from msgbus import TK_POS_LISTUSERS, FM_PARAM, TK_SYS_NAK, MBEasyContext
from mw_helper import show_filtered_list_box_dialog, show_message_dialog
from old_helper import BaseRepository
from sysactions import (
    send_message,
    get_model,
    get_podtype,
    get_user_information,
    show_messagebox,
    StopAction,
    get_poslist,
)
from xml.etree import cElementTree as eTree
from userhelper import UserHelper

from .. import USECS_PER_SEC, logger

mb_context = None


def list_opened_users(pos_id):
    msg = send_message("POS%d" % int(pos_id), TK_POS_LISTUSERS, FM_PARAM, "%s" % pos_id, timeout=600 * USECS_PER_SEC)
    if msg.token == TK_SYS_NAK:
        logger.error("Logoff POS %d - Error listing logged operators" % int(pos_id))
        raise Exception("Error listing logged operators")

    model = get_model(pos_id)
    working_mode = model.find("WorkingMode").get("usrCtrlType")

    xml = eTree.XML(msg.data)
    opened_users = [tag for tag in xml.findall("User") if not tag.get("closeTime")]
    working_users = [tag for tag in opened_users if tag.get("userControlType") == working_mode]

    if working_mode == 'QSR':
        working_users = [tag for tag in working_users if tag.get("posId") == pos_id]

    return working_users


def list_all_opened_users(pos_id):
    msg = send_message("POS%d" % int(pos_id), TK_POS_LISTUSERS, FM_PARAM, "%s" % pos_id, timeout=600 * USECS_PER_SEC)

    if msg.token == TK_SYS_NAK:
        logger.error("Logoff POS %d - Error listing logged operators" % int(pos_id))
        raise Exception("Error listing logged operators")

    xml = eTree.XML(msg.data)
    opened_users = [tag for tag in xml.findall("User") if not tag.get("closeTime")]
    working_users = [tag for tag in opened_users]

    return working_users


def list_ids_opened_users():
    with UserHelper(mb_context) as user_helper:
        users_id = []
        users = user_helper.get_all_logged_users()
        for x in users:
            users_id.append(int(x))

        return users_id


def check_valid_user(pos_id, user_id):
    user_xml = get_user_information(user_id)
    if user_xml is None:
        show_message_dialog(pos_id, message="$INVALID_USER_OR_PASSWORD")
        return False

    user_list = eTree.XML(user_xml).findall("user")
    user_filter = filter(lambda x: x.get("UserId") == str(user_id), user_list)
    if user_filter in [None, [], ""]:
        show_message_dialog(pos_id, message="$INVALID_USER_OR_PASSWORD")
        return False

    return True


def get_opened_users(pos_id):
    parsed_users_options = []
    pos_user_list = []
    is_ts_listed = False
    for pos_id_to_list in sorted(get_poslist()):
        opened_users = list_opened_users(pos_id=str(pos_id_to_list),)

        if len(opened_users) > 0:
            opened_users.sort(key=lambda x: x.get("id"))
            model = get_model(posid=pos_id_to_list,)
            pod_type = get_podtype(model=model,)
            if pod_type == "TT":
                continue

            working_mode = model.find("WorkingMode").get("usrCtrlType")

            prefix = "POS{:02d}"
            if working_mode == "TS":
                if is_ts_listed:
                    continue
                is_ts_listed = True
                prefix = "TS"
            else:
                prefix = prefix.format(pos_id_to_list)

            for user in opened_users:
                parsed_users_options.append("{} - {} - {}".format(prefix, user.get('id'), user.get('name')))
                pos_user_list.append((pos_id_to_list, user))

    if len(parsed_users_options) == 0:
        show_message_dialog(pos_id, message="$THERE_ARE_NO_OPENED_USERS")
        return None, None, None

    index = show_filtered_list_box_dialog(
        pos_id=pos_id,
        options="|".join(parsed_users_options),
        title="$LOGOUT",
        message="",
        mask="NOFILTER",
    )
    if index is None:
        return None, None, None

    pos_to_close_user = pos_user_list[index][0]
    model = get_model(posid=pos_to_close_user)
    user = pos_user_list[index][1]
    operator = [op for op in model.findall("Operators/Operator") if op.get("id") == user.get("id")]
    return pos_to_close_user, user, operator[0]




def is_user_already_logged(pos_id, user_id, try_fix=True):
    params = "\0".join(["%s" % pos_id, "true"])
    msg = send_message("POS%d" % int(pos_id), TK_POS_LISTUSERS, FM_PARAM, params, timeout=600 * USECS_PER_SEC)
    if msg.token == TK_SYS_NAK:
        logger.error("Logoff POS %d - Error getting logged users" % int(pos_id))
        return True

    xml = eTree.XML(msg.data)
    opened_users = [tag for tag in xml.findall("User") if not tag.get("closeTime")]
    working_users = [tag for tag in opened_users if tag.get("id") == user_id]

    if len(working_users) > 0:
        model = get_model(pos_id)
        _user_has_same_working_mode(model, opened_users, pos_id, user_id)

        operator = [op for op in model.findall("Operators/Operator") if op.get("id") == user_id]
        if len(operator) > 0 and operator[0].get('state') == 'PAUSED':
            return False

        user_info = [x for x in opened_users if x.get("id") == user_id][0]
        old_login = model.find("PosState").get("period") != user_info.get("sessionId").split(',period=')[-1]
        if try_fix and old_login:
            _fix_opened_users(pos_id, user_id)
            return is_user_already_logged(pos_id, user_id, False)

        show_message_dialog(pos_id, message="$OPERATOR_ALREADY_LOGGED|{}".format(user_info.get("posId")))
        return True

    return False


def _user_has_same_working_mode(model, opened_users, pos_id, user_id):
    user_info = [x for x in opened_users if x.get("id") == user_id]
    if user_info:
        pos_working_mode = model.find("WorkingMode").get("usrCtrlType")
        operator_working_mode = "QSR" if "pos" in user_info[0].get("sessionId") else "TS"
        if pos_working_mode != operator_working_mode:
            show_messagebox(pos_id, "$USER_LOGGED_IN_ANOTHER_POS_TYPE|{}".format(user_info[0].get("posId")))
            raise StopAction()


def _fix_opened_users(pos_id, user_id):
    # type: (str) -> None
    def inner_func(conn):
        # type: (Connection) -> None
        query = _fix_opened_users_query.format(user_id)
        conn.query(query)

    repository = BaseRepository(mb_context)
    repository.execute_with_connection(inner_func, pos_id)


def set_context(context):
    global mb_context
    mb_context = context


_fix_opened_users_query = """
    UPDATE UserCount SET CloseTime = (
        SELECT MAX(OpenTime) FROM UserCount WHERE OperatorId = '{0}'
    ) WHERE OperatorId = '{0}' AND CloseTime is NULL AND OpenTime < (
        SELECT MAX(CloseTime) FROM UserCount WHERE OperatorId = '{0}'
    )
"""




from abc import ABCMeta
from datetime import (
    datetime,
    timedelta,
)
from xml.etree import cElementTree as eTree
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)

from helper import round_half_away_from_zero

from application.model.benefit import (
    Benefit,
    OrderBenefitInfo,
    Conditions,
    ActionTypes,
    Action,
)
from application.model.exceptions import (
    ApplyingDiscountUnknownException,
    BenefitExpiredOrInactiveException,
    CustomBenefitAlreadyAppliedException,
    ErrorApplyingCoupon,
    ExcessiveSaleAmountException,
    InsufficientSaleAmountException,
    InvalidBrandException,
    InvalidMinSeatsException,
    InvalidSaleTypeException,
    InvalidSectorException,
    InvalidServicePeriodOrderException,
    InvalidServiceTypeException,
    MissingItemsException,
    PodTypeNotValidException,
    UserLevelNotAllowedException,
)
from application.model.order import (
    Order,
    SaleLine,
    SaleTypes,
)
from application.util import get_service_period
from iso8601 import iso8601
from typing import (
    Dict,
    List,
    Optional,
)


class BenefitValidator:
    __metaclass__ = ABCMeta

    @staticmethod
    def validate_benefits(uow, applied_benefits, burnt_items, product_groups):
        BenefitValidator.validate_benefit_static_info(applied_benefits, uow.benefit)
        BenefitValidator.validate_benefit_model_info(
            benefit=uow.benefit,
            model=uow.model,
            user_auth_level=uow.user_auth_level
        )
        BenefitValidator.validate_benefit_order_info(uow.benefit, uow.order, burnt_items, product_groups)
        BenefitValidator.validate_benefit_table_info(uow.benefit, uow.order)
        BenefitValidator.validate_benefit_actions(uow.benefit, uow.order, product_groups)

    @staticmethod
    def validate_benefit_static_info(applied_benefits, benefit):
        # type: (list, Benefit) -> None

        conditions = benefit.conditions
        BenefitValidator._is_benefit_valid_now(conditions)
        BenefitValidator._validate_unique_for_sale(applied_benefits, conditions)
        BenefitValidator._validate_single_use(benefit, applied_benefits)
        BenefitValidator._validate_unique_for_group(
            benefit=benefit,
            applied_benefits=applied_benefits,
            conditions=conditions,
        )

    @staticmethod
    def validate_benefit_model_info(
        benefit,            # type: Benefit
        model,              # type: eTree.Element
        user_auth_level     # type: bool
    ):
        # type: (...) -> None

        conditions = benefit.conditions
        BenefitValidator._is_benefit_valid_for_business_period(conditions, model)
        BenefitValidator._validate_pod_type(conditions, model)
        BenefitValidator._validate_user_min_level(conditions, user_auth_level)

    @staticmethod
    def validate_benefit_order_info(benefit, order, burnt_items, product_groups):
        # type: (Benefit, Order, Dict, Dict) -> None

        conditions = benefit.conditions
        BenefitValidator._validate_sale_type(conditions, order)
        BenefitValidator._validate_sector(conditions, order)
        BenefitValidator._validate_min_sale_amount(conditions, order)
        BenefitValidator._validate_max_sale_amount(conditions, order)
        BenefitValidator._validate_required_items_on_sale(benefit, burnt_items, order)
        BenefitValidator._validate_required_groups_on_sale(benefit, burnt_items, order, product_groups)
        BenefitValidator._validate_amount_discount(benefit, order)
        BenefitValidator._validate_service_type(conditions, order)
        BenefitValidator._validate_service_period(conditions, order)
        BenefitValidator._validate_brand(conditions, order)

    @staticmethod
    def validate_benefit_table_info(benefit, order):
        # type: (Benefit, Order) -> None

        conditions = benefit.conditions
        BenefitValidator._validate_min_seats(conditions, order)

    @staticmethod
    def validate_benefit_actions(benefit, order, product_groups):
        for action in benefit.actions:
            order_seats = BenefitValidator._get_order_seats(order=order)
            seat = action.seat

            if not seat and not isinstance(action.value, (int, float)):
                seat = action.value[0].get("seat", None)
            if (seat and not order_seats) or (seat and order_seats and int(seat) > order_seats):
                raise ErrorApplyingCoupon("NDISCOUNT_ERROR_BENEFIT_MISSING_SEAT")
            if benefit.ask_value and action.value <= 0:
                raise ErrorApplyingCoupon("NDISCOUNT_ERROR_BENEFIT_MISSING_VALUE")
            if action.name == "itemsToAdd":
                continue

            if isinstance(action.value, list):
                for value in action.value:
                    if "item" in value:
                        required_item_id = value["item"]
                        for sale_line in order.sale_lines:
                            if int(sale_line.item_id) == int(required_item_id) and sale_line.qty > 0:
                                break
                        else:
                            raise MissingItemsException('NDISCOUNT_ERROR_REQUIRED_ITEMS_ON_SALE')
                    elif "group" in value:
                        required_group = value["group"]
                        if required_group not in product_groups:
                            raise MissingItemsException('NDISCOUNT_ERROR_REQUIRED_ITEMS_ON_SALE')

                        for sale_line in order.sale_lines:
                            if int(sale_line.item_id) in product_groups[required_group] and sale_line.qty > 0:
                                break
                        else:
                            raise MissingItemsException('NDISCOUNT_ERROR_REQUIRED_ITEMS_ON_SALE')

    @staticmethod
    def _validate_required_items_on_sale(benefit, burnt_items, order):
        if not BenefitValidator._enough_items_on_order(benefit, order, burnt_items):
            raise MissingItemsException('NDISCOUNT_ERROR_REQUIRED_ITEMS_ON_SALE')

    @staticmethod
    def _validate_required_groups_on_sale(benefit, burnt_items, order, product_groups):
        if not BenefitValidator._enough_items_on_order(
            benefit=benefit,
            order=order,
            burnt_items=burnt_items,
            by_groups=True,
            product_groups=product_groups,
        ):
            raise MissingItemsException('NDISCOUNT_ERROR_REQUIRED_GROUPS_ON_SALE')

    @staticmethod
    def _validate_max_sale_amount(conditions, order):
        order_xml = order.order_picture
        due_amount = float(order_xml.get("dueAmount", 0))
        if conditions.max_sale_amount is not None and due_amount > conditions.max_sale_amount:
            raise ExcessiveSaleAmountException('NDISCOUNT_ERROR_MAX_SALE_AMOUNT')

    @staticmethod
    def _validate_min_sale_amount(conditions, order):
        order_xml = order.order_picture
        due_amount = float(order_xml.get("dueAmount", 0))
        if conditions.min_sale_amount is not None and due_amount < conditions.min_sale_amount:
            raise InsufficientSaleAmountException('NDISCOUNT_ERROR_MIN_SALE_AMOUNT')

    @staticmethod
    def _validate_unique_for_sale(applied_benefits, conditions):
        # type: (list, Conditions) -> None

        if len(applied_benefits) > 0:
            if conditions.unique_for_sale:
                raise CustomBenefitAlreadyAppliedException('NDISCOUNT_ERROR_UNIQUE_FOR_SALE')
            for applied_benefit in applied_benefits:
                benefit = Benefit(applied_benefit.get("benefit"))
                if benefit.conditions.unique_for_sale:
                    raise CustomBenefitAlreadyAppliedException('NDISCOUNT_ERROR_UNIQUE_FOR_SALE_HAS_EXISTS')

    @staticmethod
    def _is_benefit_valid_now(conditions):
        # type: (Conditions) -> None

        valid_hours = conditions.valid_hours
        now = datetime.now().isoformat().split("T")[1]
        if valid_hours.start_hour is not None and now < valid_hours.start_hour:
            raise BenefitExpiredOrInactiveException('NDISCOUNT_START_HOUR')
        if valid_hours.end_hour is not None and now > valid_hours.end_hour:
            raise BenefitExpiredOrInactiveException('NDISCOUNT_END_HOUR')

    @staticmethod
    def _is_benefit_valid_for_business_period(
        conditions,     # type: Conditions
        model,          # type: eTree.XML
    ):
        # type: (...) -> None

        period = model.find("PosState").get("period")
        datetime_period = iso8601.parse_date(period)

        if conditions.valid_days is not None and str(datetime_period.isoweekday() % 7 + 1) not in conditions.valid_days:
            raise BenefitExpiredOrInactiveException('NDISCOUNT_VALID_DAYS')

        if conditions.valid_start_period is not None and datetime_period < conditions.valid_start_period:
            raise BenefitExpiredOrInactiveException('NDISCOUNT_START_PERIOD')

        period_end = datetime_period + timedelta(days=1)
        if conditions.valid_end_period is not None and period_end > conditions.valid_end_period:
            raise BenefitExpiredOrInactiveException('NDISCOUNT_END_PERIOD')

    @staticmethod
    def _required_and_available_on_sale_from_conditions(
        conditions,     # type: Conditions
        by_groups,      # type: bool
    ):
        # type: (...) -> Tuple[Dict, Dict]

        required_on_sale = {}
        available_on_sale = {}

        if by_groups:
            for group in conditions.required_groups_on_sale:
                required_on_sale[group.group] = group.quantity
                available_on_sale[group.group] = 0
        else:
            for item in conditions.required_items_on_sale:
                key = "{};{}".format(item.context if item.context else 'any', str(item.item_id))
                required_on_sale[key] = item.quantity
                available_on_sale[key] = 0

        return required_on_sale, available_on_sale

    @staticmethod
    def _update_available_on_sale(
        required_on_sale,   # type: Dict
        available_on_sale,  # type: Dict
        sale_line,          # type: SaleLine
        quantity,           # type: float
        by_groups,          # type: bool
        product_groups,     # type: Dict
    ):
        # type: (...) -> None

        if by_groups:
            for group_key in required_on_sale.keys():
                if int(sale_line.item_id) in product_groups[group_key]:
                    available_on_sale[group_key] += quantity
            return

        context = sale_line.context
        for item_key in required_on_sale.keys():
            required_item_context, required_item_id = item_key.split(';')
            if int(sale_line.item_id) != int(required_item_id):
                continue

            if required_item_context == 'any' or required_item_context in context:
                available_on_sale[item_key] += quantity

    @staticmethod
    def _enough_items_on_order(
        benefit,                # type: Benefit
        order,                  # type: Order
        burnt_items,            # type: Dict
        by_groups=False,        # type: Optional[bool]
        product_groups=None,    # type: Dict
    ):
        # type: (...) -> int

        sale_lines = order.sale_lines
        conditions = benefit.conditions

        required_on_sale, available_on_sale = BenefitValidator._required_and_available_on_sale_from_conditions(
            conditions=conditions,
            by_groups=by_groups,
        )

        if not required_on_sale:
            return True

        for sale_line in sale_lines:
            available_quantity = sale_line.qty
            available_quantity -= 0 if conditions.ignore_burned_items else int(burnt_items.get(sale_line.key, 0))
            BenefitValidator._update_available_on_sale(
                required_on_sale=required_on_sale,
                available_on_sale=available_on_sale,
                sale_line=sale_line,
                quantity=available_quantity,
                by_groups=by_groups,
                product_groups=product_groups,
            )

        times = min([available_on_sale[i] / required_on_sale[i] for i in required_on_sale.keys()])

        times = round_half_away_from_zero(
            number=times,
            precision=3,
        )
        if times >= 1:
            benefit.times_to_apply = int(times) if conditions.burn_items else None
            return True

        return False

    @staticmethod
    def validate_added_sale_line_benefit(applied_benefit, order):
        # type: (OrderBenefitInfo, Order) -> bool

        for sale_line in order.sale_lines:
            item_removed = int(sale_line.qty) == int(sale_line.level) == 0
            benefits_exists = int(sale_line.line_number) in applied_benefit.added_sale_lines
            if benefits_exists and item_removed:
                return False

        return True

    @staticmethod
    def _validate_pod_type(conditions, model):
        # type: (Conditions, eTree.XML) -> None

        if conditions.pod_type:
            try:
                pod_type = model.find(".//WorkingMode").get("podType")
                if pod_type is None:
                    raise PodTypeNotValidException('NDISCOUNT_ERROR_POD_TYPE')

                pod_types = [pod_type]
                if pod_type.upper() == "FL":
                    pod_types = ["FC", "TS"]

                if not any(x in pod_types for x in conditions.pod_type):
                    raise PodTypeNotValidException('NDISCOUNT_ERROR_POD_TYPE')
            except PodTypeNotValidException as _:
                raise
            except Exception as _:
                raise ApplyingDiscountUnknownException('NDISCOUNT_ERROR_UNKNOWN')

    @staticmethod
    def _validate_user_min_level(conditions, user_auth_level):
        # type: (Conditions, int) -> None

        if conditions.user_min_level is not None:
            try:
                if user_auth_level < conditions.user_min_level:
                    raise UserLevelNotAllowedException('NDISCOUNT_ERROR_USER_MIN_LEVEL')
            except UserLevelNotAllowedException as _:
                raise
            except Exception as _:
                raise ApplyingDiscountUnknownException('NDISCOUNT_ERROR_UNKNOWN')

    @staticmethod
    def _validate_service_type(conditions, order):
        # type: (Conditions, Order) -> None

        if conditions.service_type:
            try:
                service_type = "Table" if order.custom_properties.get("TABLE_ID") is not None else "Quick"
                if service_type not in conditions.service_type:
                    raise InvalidServiceTypeException("NDISCOUNT_ERROR_SERVICE_TYPE_" + service_type.upper())
            except InvalidServiceTypeException as _:
                raise
            except Exception as _:
                raise ApplyingDiscountUnknownException('NDISCOUNT_ERROR_UNKNOWN')

    @staticmethod
    def _validate_sale_type(conditions, order):
        # type: (Conditions, Order) -> None

        sale_types = conditions.sale_type
        if sale_types:
            try:
                order_sale_type = SaleTypes[order.sale_type].value if SaleTypes[order.sale_type] else 'ERROR'
                any_table_order_has_correct_sale_type = BenefitValidator._validate_table_sale_type(sale_types, order)

                if order_sale_type not in sale_types and not any_table_order_has_correct_sale_type:
                    raise InvalidSaleTypeException("NDISCOUNT_ERROR_SALE_TYPE")
            except InvalidSaleTypeException as _:
                raise
            except Exception as _:
                raise ApplyingDiscountUnknownException('NDISCOUNT_ERROR_UNKNOWN')

    @staticmethod
    def _validate_sector(conditions, order):
        # type: (Conditions, Order) -> None

        if conditions.sector:
            order.get_table_info()
            try:
                if order.sector is None:
                    raise InvalidSectorException("NDISCOUNT_ERROR_SECTOR_NOT_TABLE")

                if order.sector not in conditions.sector:
                    raise InvalidSectorException("NDISCOUNT_ERROR_SECTOR")
            except InvalidSectorException as _:
                raise
            except Exception as _:
                raise ApplyingDiscountUnknownException()

    @staticmethod
    def _validate_min_seats(conditions, order):
        # type: (Conditions, Order) -> None

        if conditions.min_seats:
            order.get_table_info()
            message = 'NDISCOUNT_ERROR_MIN_REQUIRED_SEATS|{0}'.format(conditions.min_seats)
            try:
                if conditions.min_seats > order.seats:
                    raise InvalidMinSeatsException(message)
            except InvalidMinSeatsException as _:
                raise
            except Exception as _:
                raise ApplyingDiscountUnknownException('NDISCOUNT_ERROR_UNKNOWN')

    @staticmethod
    def _validate_amount_discount(benefit_to_apply, order):
        # type: (Benefit, Order) -> None

        for action in benefit_to_apply.actions:
            if action.name == ActionTypes.discountAmountOnTotal.value:
                BenefitValidator.get_discount_on_total(
                    action=action,
                    order=order,
                    benefit=benefit_to_apply,
                )
            if action.name == ActionTypes.discountPercentageOnTotal.value:
                BenefitValidator.get_discount_on_total(
                    action=action,
                    order=order,
                    benefit=benefit_to_apply,
                    use_percentage=True,
                )

    @staticmethod
    def get_discount_on_total(
        action,                # type: Action
        order,                 # type: Order
        benefit,               # type: Benefit
        use_percentage=False,  # type: bool
    ):
        # type: (...) -> None

        discount = order.total_gross * (float(action.value) / 100) if use_percentage else float(action.value)
        total_discounts = discount + order.discount_amount
        max_discount_amount = order.total_gross - 0.01
        max_value_condition = benefit.conditions.discount_max_value
        message = None

        if max_value_condition is not None and total_discounts > max_value_condition:
            remaining_discount_amount = max_value_condition - order.discount_amount
            if remaining_discount_amount <= 0:
                message = 'NDISCOUNT_MAX_DISCOUNT_APPLIED'
            else:
                if use_percentage:
                    return
                message = 'NDISCOUNT_MAX_DISCOUNT_VALUE|{0:.2f}|{1:.2f}'.format(
                    max_value_condition,
                    remaining_discount_amount,
                )
        elif action.max_value is not None and discount > action.max_value:
            message = 'NDISCOUNT_LIMITED_DISCOUNT|{0:.2f}'.format(action.max_value)
        elif discount >= order.total_gross:
            message_value = 'NDISCOUNT_ERROR_MIN_REQUIRED_SALE_AMOUNT|{0:.2f}'.format(max_discount_amount)
            message_percentage = 'NDISCOUNT_ERROR_MAX_DISCOUNT_SALE_AMOUNT|{0:.0f}'.format(action.value)
            message = message_percentage if use_percentage else message_value

        if message is not None:
            raise InsufficientSaleAmountException(message)

    @staticmethod
    def _validate_single_use(benefit, applied_benefits):
        # type: (Benefit, list) -> None

        if benefit.conditions.single_use:
            for applied_benefit in applied_benefits:
                if applied_benefit["benefit"]["id"] == benefit.id:
                    raise CustomBenefitAlreadyAppliedException('NDISCOUNT_ERROR_SINGLE_USE')


    @staticmethod
    def _validate_service_period(
        conditions,     # type: Conditions
        order           # type: Order
    ):
        # type: (...) -> None

        if conditions.service_period and (order.order_state == 'IN_PROGRESS' or not order.table_id):
            current_service_period = get_service_period(order)
            current_service_period_id = current_service_period.get("servicePeriodId", None)
            if current_service_period_id not in conditions.service_period:
                raise InvalidServicePeriodOrderException('NDISCOUNT_INVALID_SERVICE_PERIOD_ORDER')

    @staticmethod
    def _validate_table_sale_type(conditions_sale_types, order):
        # type: (List[str], Order) -> bool

        if order.table_id is None:
            return False

        for sale_line in order.sale_lines:
            if sale_line.sale_type is None:
                continue
            sale_type = SaleTypes[sale_line.sale_type].value if SaleTypes[sale_line.sale_type] else 'ERROR'
            if sale_type in conditions_sale_types:
                return True
        return False

    @staticmethod
    def _validate_brand(
        conditions,     # type: Conditions
        order,          # type: Order
    ):
        # type: (...) -> bool

        if not conditions.brands:
            return True

        if order.brand in conditions.brands:
            return True

        for sale_line in order.sale_lines:
            if sale_line.brand is None:
                continue
            elif sale_line.brand in conditions.brands:
                return True

        raise InvalidBrandException("NDISCOUNT_ERROR_BRAND")

    @staticmethod
    def _get_order_seats(
        order,  # type: Order
    ):
        # type: (...) -> Optional[int]

        seats = order.custom_properties.get("TABLE_SEATS", None)
        return int(seats) if seats else None

    @staticmethod
    def _validate_unique_for_group(
        benefit,            # type: Benefit
        applied_benefits,   # type: list
        conditions,         # type: Conditions
    ):
        # type: (...) -> None

        if len(applied_benefits) > 0:
            applied_benefits = [Benefit(x.get("benefit")) for x in applied_benefits if x.get("benefit")]
            unique_for_group = conditions.unique_for_group
            for current_benefit in applied_benefits:
                current_unique_for_group = unique_for_group or current_benefit.conditions.unique_for_group
                if current_benefit.group == benefit.group and current_unique_for_group:
                    raise CustomBenefitAlreadyAppliedException('NDISCOUNT_ERROR_UNIQUE_FOR_GROUP_HAS_EXISTS')



from typing import List


class OperatorGoals(object):
    def __init__(
        self,
        operator_id,        # type: str
        amount_goal,        # type: float
        amount_sold,        # type: float
        average_ticket,     # type: float
        item_goals,         # type: List
        tickets,            # type: float
    ):
        # type (...) -> None

        self.id = operator_id
        self.amount_goal = amount_goal
        self.amount_sold = amount_sold
        self.average_ticket = average_ticket
        self.item_goals = item_goals
        self.tickets = tickets



from application.model._AsyncUuidMessageProcessor import AsyncUuidMessageProcessor
from application.model._BaseThread import BaseThread
from application.model._BenefitAppliers import BenefitAppliers
from application.model._ListenedEvents import ListenedEvents
from application.model._ListenedTokens import ListenedTokens
from application.model._NotMappedEvents import NotMappedEvents
from application.model._OperationDescription import OperationDescription
from application.model._OperationStatus import OperationStatus
from application.model._OperationDescription import OperationDescription
from application.model._BenefitControllerDto import BenefitControllerDto
from application.model._SentTokens import LoyaltyTokens, NDiscountTokens
from application.model._DefaultBenefitApplierRepository import DefaultBenefitApplierRepository



import time

import requests
from application.configurator import Configurator
from application.model import EventData
from syserrors import SE_BADRESPONSE, SE_SUCCESS, SE_CONNECT
from typing import Optional


class EventSender(object):

    def __init__(self, configurations):
        # type: (Configurator) -> None

        self.configs = configurations
        self.logger = self.configs.logger

    def send_request(self, event_data, payload, include_store_id=False):
        # type: (EventData, str, Optional[bool]) -> int

        try:
            payload = payload.encode("utf-8")
            self.logger.info("[{}] Sending event to server. Payload:\n{}".format(event_data.type, payload))

            headers = {
                    "Accept": "application/json",
                    "Content-type": "application/json",
                    "x-api-key": self.configs.api_key
            }

            post_url = self.configs.server_url + self.configs.views.get(event_data.type)
            if include_store_id:
                post_url += "/{}".format(self.configs.store_id)

            response = requests.post(post_url, headers=headers, data=payload, timeout=30, verify=False)  # NOSONAR

            if response.status_code != 200:
                log_message = "[{0}] Error sending event to server. Status: {1} - {2}"
                self.logger.info(log_message.format(event_data.type, response.status_code, response.content))
                time.sleep(self.configs.retry_wait / 1000.0)
                return SE_BADRESPONSE

            self.logger.info("[{}] Success sending event to server".format(event_data.type))
            return SE_SUCCESS

        except Exception as _:
            message = "[{}] Communication error sending event to server".format(event_data.type)
            self.logger.exception(message)
            return SE_CONNECT



u"""
Fixer for:
it.__next__() -> it.next().
next(it) -> it.next().
"""

from lib2to3.pgen2 import token
from lib2to3.pygram import python_symbols as syms
from lib2to3 import fixer_base
from lib2to3.fixer_util import Name, Call, find_binding, Attr

bind_warning = u"Calls to builtin next() possibly shadowed by global binding"


class FixNext(fixer_base.BaseFix):

    PATTERN = u"""
    power< base=any+ trailer< '.' attr='__next__' > any* >
    |
    power< head='next' trailer< '(' arg=any ')' > any* >
    |
    classdef< 'class' base=any+ ':'
              suite< any*
                     funcdef< 'def'
                              attr='__next__'
                              parameters< '(' NAME ')' > any+ >
                     any* > >
    """

    def transform(self, node, results):
        assert results

        base = results.get(u"base")
        attr = results.get(u"attr")
        head = results.get(u"head")
        arg_ = results.get(u"arg")
        if arg_:
            arg = arg_.clone()
            head.replace(Attr(Name(unicode(arg),prefix=head.prefix),
                              Name(u"next")))
            arg_.remove()
        elif base:
            attr.replace(Name(u"next", prefix=attr.prefix))



from typing import Optional

from httpmessageprocessorutil.automapped import auto_map


@auto_map({"system_api_key": {"name": "X-Api-Key", "location": "header", "type": str},
           "pos_id": {"name": "posId", "type": str},
           "print_coupon": {"name": "printCoupon", "type": Optional[str]}},
          )
class OrderSaveRequest(object):
    def __init__(
        self,
        system_api_key,           # type: str
        pos_id,                   # type: str
        print_coupon='True',      # type: Optional[str]
    ):
        # type: (...) -> None

        self.system_api_key = system_api_key
        self.pos_id = pos_id
        self.print_coupon = True if print_coupon is None else print_coupon.lower() == 'true'



# encoding: utf-8
from _JustificationDiscountBodyDto import JustificationDiscountBodyDto
from _JustificationDiscountHeaderDto import JustificationDiscountHeaderDto


class JustificationDiscountGeneratorDto(object):
    def __init__(
        self,
        header,      # type: JustificationDiscountHeaderDto
        body,        # type: JustificationDiscountBodyDto
    ):
        # type: (...) -> None

        self.header = header
        self.body = body



# encoding: utf-8
from old_helper import BaseRepository


class DeliveryOrderRepository(BaseRepository):

    def __init__(self, mb_context, pos_list):
        super(DeliveryOrderRepository, self).__init__(mb_context)
        self.pos_list = pos_list

    def get_order_json(self, order_id):
        def inner_func(conn):
            query = self._OrdersQuery.format(order_id)

            for x in conn.select(query):
                return x.get_entry("Value")

            return ''

        return self.execute_parallel_in_all_databases(inner_func, self.pos_list)

    _OrdersQuery = """
                    SELECT Value
                    FROM OrderCustomProperties
                    WHERE OrderId = '{}' AND (Key = 'REMOTE_ORDER_JSON' or Key = 'MANUAL_DELIVERY_ORDER_JSON')
                  """



from reportsdata.samplereport import SampleReportGenerator, SampleReportPresenter


class ReportGenerator(object):
    @staticmethod
    def generate(mb_context, pos_id, param1, param2, param3, param4):
        return SampleReportGenerator(), SampleReportPresenter()



import collections
import json
import sys

from json import loads
from threading import Lock
from typing import Optional, List, Dict

from persistence import Connection
from mwhelper import BaseRepository
from mbcontextmessagehandler import MbContextMessageBus

from application.model import Configurations
from application.util import get_range_discount_ids


class NDiscountRepository(BaseRepository):

    def __init__(self, configs, message_bus):
        # type: (Configurations, MbContextMessageBus) -> None
        super(NDiscountRepository, self).__init__(message_bus.mbcontext)

        self.logger = configs.logger
        self.configs = configs
        self.message_bus = message_bus
        self.product_cache = []
        self.product_group_cache = {}
        self.service_periods_price_keys_cache = {}
        self.service_periods_cache = {}
        self.automatic_benefits = []
        self.popup_lock = Lock()

    def build_singletons(self):
        self._create_discount_id_range()
        self._get_automatic_benefits()
        self._get_all_products()
        self._get_all_product_groups()
        self._get_service_periods_price_keys()
        self._get_service_periods()

    def get_benefits_by_visibility_from_database(self, only_visible):
        # type: (bool) -> List[dict]

        def inner_func(conn):
            # type: (Connection) -> List[dict]

            query = "SELECT BenefitId, BenefitJSON FROM StoredBenefits"
            cursor = conn.select(query)

            all_benefits = []
            for row in cursor:
                benefit_item = loads(row.get_entry(1))
                if not only_visible or benefit_item.get("visible", False):
                    all_benefits.append(dict(benefitId=row.get_entry(0), benefitJson=benefit_item))

            return all_benefits

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def _get_all_products(self):
        # type: () -> None

        def inner_func(conn):
            # type: (Connection) -> List[int]

            query = "SELECT ProductCode FROM Product"
            cursor = conn.select(query)

            all_products = []
            for row in cursor:
                product_item = int(row.get_entry(0))
                all_products.append(product_item)

            return all_products

        self.product_cache = self.execute_with_connection(inner_func, service_name="Persistence")

    def _get_service_periods_price_keys(self):
        # type: () -> None

        def inner_func(conn):
            # type: (Connection) -> Dict

            query = "SELECT PriceKey, PriceListId FROM Price"
            cursor = conn.select(query)

            all_service_periods_price_keys = {}
            for row in cursor:
                price_key, price_list = int(row.get_entry(0)), str(row.get_entry(1))
                all_service_periods_price_keys[price_key] = price_list

            return all_service_periods_price_keys

        self.service_periods_price_keys_cache = self.execute_with_connection(inner_func, service_name="Persistence")

    def _get_service_periods(self):
        # type: () -> None

        def inner_func(conn):
            # type: (Connection) -> Dict

            query = "SELECT ServicePeriodId, PriceListId FROM ServicePeriod"
            cursor = conn.select(query)

            all_service_periods = {}
            for row in cursor:
                service_period_id, price_list = int(row.get_entry(0)), str(row.get_entry(1))
                all_service_periods[service_period_id] = price_list

            return all_service_periods

        self.service_periods_cache = self.execute_with_connection(inner_func, service_name="Persistence")

    def _get_all_product_groups(self):
        # type: () -> None

        def inner_func(conn):
            # type: (Connection) -> Dict

            query = """SELECT 
                           ProductCode, CustomParamValue 
                       FROM ProductCustomParams 
                       WHERE CustomParamId = 'PLUFamilyGroup'"""
            cursor = conn.select(query)

            all_product_groups = {}
            for row in cursor:
                product = int(row.get_entry(0))
                groups = str(row.get_entry(1)).split(';')
                for g in groups:
                    if g in all_product_groups:
                        all_product_groups[g].append(product)
                    else:
                        all_product_groups[g] = [product]

            return all_product_groups

        self.product_group_cache = self.execute_with_connection(inner_func, service_name="Persistence")

    def _create_discount_id_range(self):
        # type: () -> None

        def inner_func(conn):
            # type: (Connection) -> List[int]

            query = "SELECT DiscountId FROM Discounts"
            cursor = conn.select(query)

            discounts_id = [int(row.get_entry(0)) for row in cursor]
            range_discount_ids = get_range_discount_ids()
            for range_id in range_discount_ids:
                if range_id not in discounts_id:
                    query = """INSERT OR REPLACE INTO Discounts (DiscountId, DiscountDescr, Script, ValidFrom) {0}"""
                    query = query.format("VALUES ({0}, '', '', '2010-04-29')")
                    query = query.format(range_id)
                    conn.select(query)
                    discounts_id.append(range_id)

            return discounts_id

        return self.execute_with_connection(inner_func, service_name="Persistence")

    def _get_automatic_benefits(self):
        # type: () -> None

        def inner_func(conn):
            # type: (Connection) -> List[str]

            query = "SELECT BenefitId, BenefitJSON FROM StoredBenefits"
            cursor = conn.select(query)

            discounts = []
            benefits = {}
            for row in cursor:
                if row.get_entry(1) is None or "automatic" not in row.get_entry(1):
                    continue

                data = json.loads(row.get_entry(1))
                if data["automatic"] in ["ON_TOTAL", "ON_SALE", "ON_TABLE_OPEN", "ON_TAB_OPEN"]:
                    benefit = json.dumps(dict(benefitId=row.get_entry(0), benefit=data))
                    benefit_priority = data.get("priority", sys.maxsize)
                    if benefit_priority not in benefits:
                        benefits[benefit_priority] = []
                    benefits[benefit_priority].append(benefit)

            ordered_benefits = collections.OrderedDict(sorted(benefits.items()))
            for x in ordered_benefits.items():
                discounts.extend(x[1])
            return discounts

        self.automatic_benefits = self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def get_all_products(self):
        # type: () -> List[int]

        return self.product_cache

    def get_all_automatic_benefits(self):
        # type: () -> List[str]

        return self.automatic_benefits

    def get_all_service_periods_price_keys(self):
        # type: () -> dict

        return self.service_periods_price_keys_cache

    def get_all_service_periods(self):
        # type: () -> dict

        return self.service_periods_cache

    def get_all_product_groups(self):
        # type: () -> Dict

        return self.product_group_cache



from logging import Logger

from model.config import Config
from model.repository import PosRepositoryModel
from model.services import (
    CouponServiceModel,
    OrderServiceModel
)
from model.error_type import PosValidationErrorsType
from model.exception import ValidationError
from interactor.route.order.apply_coupon.dto import ApplyCouponRequest
from validation_util import check_if_pos_id_is_valid, validate_api_key, check_if_order_is_in_progress_or_totaled
from posot import OrderTaker
from typing import Optional

FREE_COUPON_DATA = 'FREE_COUPON_DATA'


class ApplyCouponInteractor(object):
    def __init__(
            self,
            logger,             # type: Logger
            config,             # type: Config
            pos_repository,     # type: PosRepositoryModel
            coupon_service,     # type: CouponServiceModel
            order_service       # type: OrderServiceModel
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.coupon_service = coupon_service
        self.order_service = order_service
        self.model = None
        self.pos_ot = None  # type: Optional[OrderTaker]
        self.price_list = None

    def execute(self, request):
        # type: (ApplyCouponRequest) -> None

        pos_id = validate_api_key(request, self.config.api_key_mapping)
        check_if_pos_id_is_valid(self.pos_repository, pos_id)

        coupon_data, coupon_applied = self.pos_repository.apply_coupon(
            pos_id=pos_id,
            coupon_code=request.coupon_code
        )

        self.model = self.pos_repository.get_model(pos_id)
        self.pos_ot = self.pos_repository.get_pos_ot(self.model)
        self.price_list = self.pos_repository.get_price_list(self.model)

        if self.pos_repository.has_current_order(self.model):
            check_if_order_is_in_progress_or_totaled(self.pos_repository, self.model)

        if coupon_applied:
            self.logger.info("Success coupon applied. CouponData: {}".format(coupon_data))
            order = self.pos_repository.get_current_order(self.model)
            order_id = order.get('orderId')
            order = self.pos_repository.get_order_details(
                pos_id=pos_id,
                order_id=order_id,
                pos_ot=self.pos_ot
            )
            if order:
                coupon_data = self.order_service.convert_mw_order_to_order_detail(order)

            return coupon_data

        else:
            self.logger.info("Error applying coupon. CouponData: {}".format(coupon_data))
            raise ValidationError(PosValidationErrorsType.CannotApplyCoupon, coupon_data)



from production.box._OrderChangerProductionBox import OrderChangerProductionBox


class RemoveDoingWhenDoneBox(OrderChangerProductionBox):
    def __init__(self, name, sons, logger=None):
        super(RemoveDoingWhenDoneBox, self).__init__(name, sons, logger)
        self.tags = ["doing", "done"]

    def change_order(self, order):
        for item in order.items:
            self.remove_doing_when_done_from_item(item)

        return order

    def remove_doing_when_done_from_item(self, item):
        if item.has_tag("done") and item.has_tag("doing"):
            item.remove_tag("doing")

        for son in item.items:
            self.remove_doing_when_done_from_item(son)



# encoding: utf-8

import re
from datetime import datetime

from old_helper import BaseRepository
from old_helper import convert_from_utf_to_localtime
from msgbus import MBEasyContext
from persistence import Connection
from typing import List, Optional

from ..model import Order


class OrderRepository(BaseRepository):
    TypeRealDate = "RealDate"
    TypeBusinessPeriod = "BusinessPeriod"
    TypeSessionId = "SessionId"

    def __init__(self, mbcontext, pos_list):
        # type: (MBEasyContext, List[int]) -> None
        super(OrderRepository, self).__init__(mbcontext)
        self.pos_list = pos_list

    def get_voided_orders_by_real_date(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_voided_orders(self.TypeRealDate, initial_date, end_date, operator_id, None, report_pos)

    def get_voided_orders_by_business_period(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_voided_orders(self.TypeBusinessPeriod, initial_date, end_date, operator_id, None, report_pos)

    def get_voided_orders_by_session_id(self, session_id):
        # type: (unicode) -> List[Order]
        return self._get_voided_orders(self.TypeSessionId, None, None, None, session_id, None)

    def _get_voided_orders(self, report_type, initial_date, end_date, operator_id, session_id, report_pos):
        # type: (unicode, Optional[datetime], Optional[datetime], Optional[unicode], Optional[unicode], Optional[unicode]) -> List[Order]
        def inner_func(conn):
            # type: (Connection) -> List[Order]

            try:
                if report_type == self.TypeRealDate:
                    query = self._VoidedOrdersByRealDateQuery
                    date_format = "%Y-%m-%d"
                    query = query.format(initial_date.strftime(date_format), end_date.strftime(date_format))
                elif report_type == self.TypeBusinessPeriod:
                    query = self._VoidedOrdersByBusinessPeriodQuery
                    date_format = "%Y%m%d"
                    query = query.format(initial_date.strftime(date_format), end_date.strftime(date_format))
                else:
                    query = self._VoidedOrdersBySessionIdQuery
                    query = query.format(session_id)

                if report_type == self.TypeRealDate or report_type == self.TypeBusinessPeriod and operator_id is not None:
                    query += " and o.SessionId like '%user={0},%'".format(operator_id)

                order_with_payments = [(int(x.get_entry(0)),
                                        unicode(x.get_entry(1).split(" - ")[-1]),
                                        int(re.search(r"(?:user=)(.\d*)", x.get_entry(2)).group(1)),
                                        x.get_entry(3),
                                        convert_from_utf_to_localtime(datetime.strptime(x.get_entry(4), "%Y-%m-%dT%H:%M:%S.%f")),
                                        float(x.get_entry(5))) for x in conn.select(query)]

                all_orders = []  # type: List[Order]
                for order_tuple in order_with_payments:
                    order_id = order_tuple[0]
                    order_reason = order_tuple[1]
                    order_operator_id = order_tuple[2]
                    order_authorizer = order_tuple[3]
                    order_datetime = order_tuple[4]
                    order_total = order_tuple[5]

                    order = Order(order_id, order_reason, order_operator_id, order_authorizer, order_datetime, order_total)
                    all_orders.append(order)

                return all_orders
            except:
                return list({"error": "error"}.values())

        report_pos_list = self.pos_list if report_pos is None else (report_pos, )
        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, report_pos_list)

    _VoidedOrdersByRealDateQuery = \
"""select o.OrderId, coalesce(vrs.Value, 'Nao Populado') as VoidReason, o.SessionId, aut.Value as AuthenticationUser, osh.OrderDate as Timestamp, coalesce(o.TotalGross, 0) as TotalGross
    from Orders o
    left join (
        select OrderId, OrderDate 
        from (
            select OrderId, (max(Timestamp)) OrderDate
                from (
                    select OrderId, StateId, Timestamp 
                        from OrderStateHistory 
                        where strftime('%Y-%m-%d', Timestamp, 'localtime') > strftime('%Y-%m-%d', '{0}' , '-1 day') and strftime('%Y-%m-%d', Timestamp, 'localtime') < strftime('%Y-%m-%d', '{1}' , '+1 day')
                )
                where StateId = 4
                group by OrderId
        ) a where strftime('%Y-%m-%d', a.OrderDate, 'localtime') >= '{0}' and strftime('%Y-%m-%d', a.OrderDate, 'localtime') <= '{1}'
    ) osh on o.OrderId = osh.OrderId
    left join OrderCustomProperties vrs on vrs.OrderId = o.OrderId and vrs.Key = 'VOID_REASON_DESCR'
    left join OrderCustomProperties aut on aut.OrderId = o.OrderId and aut.Key = 'AUTHENTICATION_USER'
    where o.StateId = 4 and osh.OrderDate is not null and osh.StateId = 4"""

    _VoidedOrdersByBusinessPeriodQuery = \
"""select o.OrderId, coalesce(vrs.Value, 'Nao Populado') as VoidReason, o.SessionId, aut.Value as AuthenticationUser, osh.Timestamp as Timestamp, coalesce(o.TotalGross, 0) as TotalGross
    from Orders o
	left join OrderStateHistory osh on osh.OrderId = o.OrderId
	left join OrderCustomProperties vrs on vrs.OrderId = o.OrderId and vrs.Key = 'VOID_REASON_DESCR'
	left join OrderCustomProperties aut on aut.OrderId = o.OrderId and aut.Key = 'AUTHENTICATION_USER'
	where o.BusinessPeriod >= '{0}' and o.BusinessPeriod <= '{1}' and o.StateId = 4 and osh.Timestamp is not null and osh.StateId = 4"""

    _VoidedOrdersBySessionIdQuery = \
"""select o.OrderId, coalesce(vrs.Value, 'Nao Populado') as VoidReason, o.SessionId, aut.Value as AuthenticationUser, osh.Timestamp as Timestamp, coalesce(o.TotalGross, 0) as TotalGross
	from Orders o
	left join OrderStateHistory osh on osh.OrderId = o.OrderId
	left join OrderCustomProperties vrs on vrs.OrderId = o.OrderId and vrs.Key = 'VOID_REASON_DESCR'
	left join OrderCustomProperties aut on aut.OrderId = o.OrderId and aut.Key = 'AUTHENTICATION_USER'
	where o.SessionId = '{0}' and o.StateId = 4 and osh.Timestamp is not null and osh.StateId = 4"""



import locale

from datetime import datetime
from decimal import Decimal
from typing import (
    Dict,
    List,
    Optional,
)

import iso8601

from sysactions import get_model

from actions.alerts.model import UserAlertsTable
from utilfunctions import format_currency


class ConfiguredAlertsParser(object):

    def __init__(
        self,
        pos_id,             # type: int
        configured_alerts,  # type: Dict
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.model = get_model(self.pos_id)
        self.configured_alerts = configured_alerts
        self.last_updated = self.configured_alerts.pop("last_updated")
        self.condition_formatter = {
            "cancelAmount": self._cancel_amount_fmt,
            "discountAmount": self._discount_amount_fmt,
            "discountPercentage": self._discount_percentage_fmt,
            "onlyAfterFiscalization": self._only_after_fiscalization_fmt,
        }

    def get_data_table(self):
        # type: (...) -> List[str]

        user_ids = set()
        for alert in self.configured_alerts.values():
            user_ids.add(alert.get("user", {}).get("id", "?"))

        user_ids.discard("?")
        alerts_table = []

        for user_id in user_ids:
            user_alerts = self._list_user_alerts(user_id)
            alerts_table.extend(['#'.join(row) for row in user_alerts.build_table_rows()])

        return alerts_table

    def get_last_updated(self):
        # type: (...) -> Optional[datetime]

        if self.last_updated is None:
            return None

        return iso8601.parse_date(self.last_updated)

    def _list_user_alerts(
        self,
        user_id,  # type: str
    ):
        # type: (...) -> UserAlertsTable

        user_alerts_table = None

        for alert in self.configured_alerts.values():
            user = alert.get("user", {})
            if user_id == user.get("id"):
                if user_alerts_table is None:
                    user_alerts_table = UserAlertsTable(
                        user_id=user_id,
                        user_name=user.get("name", "$UNNAMED_USER"),
                    )
                self._list_alert_conditions(
                    alert,
                    user_alerts_table,
                )

        return user_alerts_table

    def _list_alert_conditions(
        self,
        alert,              # type: Dict
        user_alerts_table,  # type: UserAlertsTable
    ):
        # type: (...) -> None

        conditions = alert.get("conditions", {})
        notify_on = "$ALERT_NOTIFY_ON_{}".format(alert.get("notifyOn", "UNKNOW_ALERT"))

        for condition, value in conditions.items():
            formatter = self.condition_formatter.get(condition)
            if formatter is None:
                condition_warn = "{} (?)".format(condition)
                fmt_value = "{} (?)".format(value)
            else:
                condition_warn = None
                fmt_value = formatter(value)

            if fmt_value is not None:
                alert_type = condition_warn or notify_on
                user_alerts_table.add(
                    alert_type,
                    fmt_value,
                )

    def _discount_amount_fmt(
        self,
        value,  # type: str
    ):
        # type: (...) -> str

        return format_currency(self.model, float(value))

    def _cancel_amount_fmt(
        self,
        value,  # type: str
    ):
        # type: (...) -> str

        return format_currency(self.model, float(value))

    @staticmethod
    def _discount_percentage_fmt(
        value,  # type: str
    ):
        # type: (...) -> str

        return _format_percent(value)

    @staticmethod
    def _only_after_fiscalization_fmt(
        _,  # type: str
    ):
        # type: (...) -> None

        return None


def _format_percent(
    value,  # type: str
):
    # type: (...) -> str

    s = locale.format_string("%.2f", Decimal(value))
    return "{}%".format(s.rstrip("0").rstrip(".,"))



import json
import logging

from typing import (
    Dict,
    Optional,
)

from bustoken import TK_MAINTENANCE_DATABASE_CLEANUP_STATUS
from mw_helper import show_message_options_dialog
from msgbus import (
    FM_STRING,
    TK_SYS_ACK,
)
from pyscripts import mbcontext as mb_context


class MaintenanceInteractor(object):
    def __init__(self):
        # type: (...) -> None

        self.logger = logging.getLogger("PosActions")  # type: logging.Logger

    def verify_clean_databases(
        self,
        period,  # type: str
        pos_id,  # type: str
    ):
        # type: (...) -> None

        status = self._get_status()
        if status:
            cleanup_required = status.get("cleanup_required", False)
            if cleanup_required:
                self._cleanup(period, pos_id)

    def _get_status(self):
        # type: (...) -> Optional[Dict[str, bool]]

        try:
            response = mb_context.MB_EasySendMessage(
                "Maintenance00",
                TK_MAINTENANCE_DATABASE_CLEANUP_STATUS,
                format=FM_STRING,
                timeout=10000 * 1000
            )
            if response.token == TK_SYS_ACK:
                status = json.loads(response.data)
                return status
            else:
                self.logger.error("cannot retrieve database cleanup status from maintenance")
                return None
        except Exception:  # noqa
            self.logger.exception("error retrieving database cleanup status from maintenance")
            return None

    def _cleanup(
        self,
        period,  # type: str
        pos_id,  # type: str
    ):
        # type: (...) -> None

        try:
            response = show_message_options_dialog(
                pos_id,
                "$YES|$NO",
                "$INFORMATION",
                "$CLOSE_DAY_DATABASE_CLEANUP_CONFIRMATION"
            )
            if response is None or response == 1:
                self.logger.info("database cleanup on close day declined by user")
                return

            event_data = {
                "posId": pos_id,
                "businessPeriod": period,
                "vacuum": True,
                "ignoreMaintenanceWindow": True
            }
            mb_context.MB_EasyEvtSend(
                subject="CleanDatabases",
                type="",
                xml=json.dumps(event_data),
                synchronous=False,
                timeout=-1
            )
        except Exception:  # noqa
            self.logger.exception("error sending clean databases event")



import json
from typing import Any

from application.interactor import GetAndLockVoucherInteractor
from application.model import AsyncUuidMessageProcessor
from application.model.configuration import Configurations
from messagebus import DefaultToken, DataType, Message


class GetAndLockVoucherProcessor(AsyncUuidMessageProcessor):
    def __init__(self, configs, interactor):
        # type: (Configurations, GetAndLockVoucherInteractor) -> None

        super(GetAndLockVoucherProcessor, self).__init__()
        self.logger = configs.logger
        self.configs = configs
        self.interactor = interactor

    def parse_data(self, data, event_type):
        # type: (str, str) -> Any

        parsed_data = json.loads(data)
        pos_id = int(parsed_data.get("posId"))
        voucher_id = str(parsed_data.get("voucherId"))
        return pos_id, voucher_id

    def call_business(self, params):
        # type: ((int, str)) -> str

        pos_id, voucher_id = params
        
        return self.interactor.get_and_lock_voucher(pos_id, voucher_id)

    def format_response(self, message_bus, message, event, request, response):
        if response:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.string, response))
        else:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



from typing import (
    Any,
    Callable,
)


class CacheManager(object):
    def __init__(
        self,
        sector_config,  # type: dict
    ):
        # type: (...) -> None

        self.sector_config = sector_config
        self.cached_objects = {}
        self.cache_key = "POSID={}|DEVICETYPE={}"

    def get_cached_object(
        self,
        sector,                             # type: str
        device_type,                        # type: str
        pos_id,                             # type: str
        method_to_obtain_cache,             # type: Callable[[Any], Any]
        method_to_obtain_cache_params,      # type: tuple
    ):
        # type: (...) -> str

        sector_key = sector.upper() if sector else sector
        device_type_key = device_type.upper()
        if sector_key in self.sector_config and device_type_key in self.sector_config[sector_key]:
            return self.sector_config[sector_key][device_type_key]
        cache_key = self.cache_key.format(pos_id, device_type_key)
        if cache_key not in self.cached_objects:
            object_to_cache = method_to_obtain_cache(*method_to_obtain_cache_params)
            self.cached_objects[cache_key] = object_to_cache

        return self.cached_objects[cache_key]



from xml.etree import cElementTree as eTree

from dailygoals_services.model.repository import DailyGoalsRepository
from dailygoals_services.service import DailyGoalsService
from externals.util import LoggerUtil
from sysactions import get_model, get_posot

logger = LoggerUtil.get_logger_name(__name__)


class UpdateTotalSoldTableInteractor:
    def __init__(self, daily_goals_repository, daily_goals_service):
        # type: (DailyGoalsRepository, DailyGoalsService) -> None

        self.daily_goals_repository = daily_goals_repository
        self.daily_goals_service = daily_goals_service

    def update(self, data):
        # type: (str) -> None

        logger.info("[handle_event] TABLE_CLOSED event received")
        table_picture = eTree.XML(data)
        payment_order = table_picture.findall('.//PaymentOrder')[0]
        pos_id = table_picture.find("Table").get("currentPOSId", None)
        order_id = payment_order.get("orderId", None)

        order = self._get_order_picture(pos_id, order_id)
        number_of_seats = 0

        if not order or order.get("state") != "PAID":
            return

        if table_picture.find("Table").get("serviceSeats") is not None:
            number_of_seats = int(table_picture.find("Table").get("serviceSeats"))

        self.daily_goals_service.update_daily_sold(order, number_of_seats=number_of_seats)

    @staticmethod
    def _get_order_picture(pos_id, order_id):
        # type: (str, str) -> eTree.Element

        order = None
        if pos_id and order_id:
            model = get_model(pos_id)
            pos_ot = get_posot(model)
            orders = eTree.XML(pos_ot.orderPicture(pos_id, order_id))
            order = orders.findall('.//Order')[0]
        return order



# -*- coding: utf-8 -*-
from sysactions import action, get_business_period, get_model, check_current_order, get_custom

from common_close_user import common_close_user
from custom_sysactions import user_authorized_to_execute_action

from .. import logger
from ..pos.check_table_status import has_tab_or_tables_opened, show_open_table_dialog


@action
def logoffuser(pos_id, *args):
    logger.debug("Logoff POS %d - Iniciando logoff." % int(pos_id))

    model = get_model(pos_id)
    check_current_order(pos_id, model=model, need_order=False)
    period = get_business_period(model)

    operator = model.find("Operator")
    if operator is None:
        return False

    user_id = operator.get("id")
    if has_tab_or_tables_opened(period, pos_id, user_id):
        show_open_table_dialog(pos_id)
        return False

    return log_off_user_close_user(pos_id, model, user_id)


@user_authorized_to_execute_action
def log_off_user_close_user(pos_id, model, user_id, auth_data=None):
    if auth_data is None:
        authorizer = get_custom(model, 'Last Manager ID')
        auth_data = authorizer

    return common_close_user(pos_id, model, user_id, authorizer=auth_data)



class GetUsedServiceRequest(object):
    def __init__(
        self,
        pos_id,         # type: str
        device_type,    # type: str
        sector,         # type: str
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.device_type = device_type
        self.sector = sector



"""
Application Profiler
====================

This module provides a middleware that profiles each request with the
:mod:`cProfile` module. This can help identify bottlenecks in your code
that may be slowing down your application.

.. autoclass:: ProfilerMiddleware

:copyright: 2007 Pallets
:license: BSD-3-Clause
"""
from __future__ import print_function

import os.path
import sys
import time
from pstats import Stats

try:
    from cProfile import Profile
except ImportError:
    from profile import Profile


class ProfilerMiddleware(object):
    """Wrap a WSGI application and profile the execution of each
    request. Responses are buffered so that timings are more exact.

    If ``stream`` is given, :class:`pstats.Stats` are written to it
    after each request. If ``profile_dir`` is given, :mod:`cProfile`
    data files are saved to that directory, one file per request.

    The filename can be customized by passing ``filename_format``. If
    it is a string, it will be formatted using :meth:`str.format` with
    the following fields available:

    -   ``{method}`` - The request method; GET, POST, etc.
    -   ``{path}`` - The request path or 'root' should one not exist.
    -   ``{elapsed}`` - The elapsed time of the request.
    -   ``{time}`` - The time of the request.

    If it is a callable, it will be called with the WSGI ``environ``
    dict and should return a filename.

    :param app: The WSGI application to wrap.
    :param stream: Write stats to this stream. Disable with ``None``.
    :param sort_by: A tuple of columns to sort stats by. See
        :meth:`pstats.Stats.sort_stats`.
    :param restrictions: A tuple of restrictions to filter stats by. See
        :meth:`pstats.Stats.print_stats`.
    :param profile_dir: Save profile data files to this directory.
    :param filename_format: Format string for profile data file names,
        or a callable returning a name. See explanation above.

    .. code-block:: python

        from werkzeug.middleware.profiler import ProfilerMiddleware
        app = ProfilerMiddleware(app)

    .. versionchanged:: 0.15
        Stats are written even if ``profile_dir`` is given, and can be
        disable by passing ``stream=None``.

    .. versionadded:: 0.15
        Added ``filename_format``.

    .. versionadded:: 0.9
        Added ``restrictions`` and ``profile_dir``.
    """

    def __init__(
        self,
        app,
        stream=sys.stdout,
        sort_by=("time", "calls"),
        restrictions=(),
        profile_dir=None,
        filename_format="{method}.{path}.{elapsed:.0f}ms.{time:.0f}.prof",
    ):
        self._app = app
        self._stream = stream
        self._sort_by = sort_by
        self._restrictions = restrictions
        self._profile_dir = profile_dir
        self._filename_format = filename_format

    def __call__(self, environ, start_response):
        response_body = []

        def catching_start_response(status, headers, exc_info=None):
            start_response(status, headers, exc_info)
            return response_body.append

        def runapp():
            app_iter = self._app(environ, catching_start_response)
            response_body.extend(app_iter)

            if hasattr(app_iter, "close"):
                app_iter.close()

        profile = Profile()
        start = time.time()
        profile.runcall(runapp)
        body = b"".join(response_body)
        elapsed = time.time() - start

        if self._profile_dir is not None:
            if callable(self._filename_format):
                filename = self._filename_format(environ)
            else:
                filename = self._filename_format.format(
                    method=environ["REQUEST_METHOD"],
                    path=(
                        environ.get("PATH_INFO").strip("/").replace("/", ".") or "root"
                    ),
                    elapsed=elapsed * 1000.0,
                    time=time.time(),
                )
            filename = os.path.join(self._profile_dir, filename)
            profile.dump_stats(filename)

        if self._stream is not None:
            stats = Stats(profile, stream=self._stream)
            stats.sort_stats(*self._sort_by)
            print("-" * 80, file=self._stream)
            print("PATH: {!r}".format(environ.get("PATH_INFO", "")), file=self._stream)
            stats.print_stats(*self._restrictions)
            print("-" * 80 + "\n", file=self._stream)

        return [body]



from logging import Logger
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)

import requests
from requests import Response

from helper import retry

from services.model import (
    Configs,
    ForbiddenError,
    InUseError,
    NotAvailableError,
    NotFound,
    NotFoundError,
    NoResponseError,
    ProcessorError,
)
from services.util import (
    get_store_id,
    get_boh_url,
    get_boh_api_key,
)


class AlertsAPIRepository(object):

    def __init__(
        self,
        configs,  # type: Configs
        logger,   # type: Logger
    ):
        # type: (...) -> None

        self.logger = logger
        self.configs = configs

    def retrieve_configured_alerts(
        self,
    ):
        # type: (...) -> Optional[str, Exception]

        try:
            url = "{}{}".format(
                get_boh_url(),
                self.configs.alerts["EndPoints"]["ConfiguredAlerts"],
            )
            url = url.format(storeId=get_store_id())
            headers = self._get_headers()
            errors_list = []

            response = retry(
                logger=self.logger,
                method_name="retrieve_configured_alerts",
                n=1,
                sleep_seconds=1,
                trying_to_do=lambda: requests.request("GET", url=url, headers=headers, timeout=20),
                is_success=lambda resp: resp.status_code == 200,
                success=lambda resp: resp,
                failed=lambda resp: errors_list.append((resp.status_code, resp))
            )
            if response:
                return response.content

            if not errors_list:
                raise NotFound()

            response_error = self._get_response_error(errors_list)
            self._raise_lock_custom_exception(
                response_error.status_code,
                response_error
            )
        except Exception as _:  # noqa
            self.logger.exception("Error retrieving configured alerts")
            raise

    @staticmethod
    def _get_headers(
    ):
        # type: (...) -> Dict

        headers = {
            "X-Api-Key": get_boh_api_key(),
            "Content-type": "application/json",
        }
        return headers

    def _raise_lock_custom_exception(
        self,
        status_code,     # type: int
        response_error,  # type: Response
    ):
        # type: (...) -> None

        if status_code == 204:
            raise NoResponseError(self.logger, response_error)
        if status_code == 403:
            raise ForbiddenError(self.logger, response_error)
        if status_code == 404:
            raise NotFoundError(self.logger, response_error)
        if status_code == 406:
            raise NotAvailableError(self.logger, response_error)
        if status_code == 409:
            raise InUseError(self.logger, response_error)
        if status_code >= 500:
            raise ProcessorError(self.logger, response_error)

    @staticmethod
    def _get_response_error(
        errors_list,  # type: List[Tuple]
    ):
        # type: (...) -> Response

        last_error = errors_list[-1]
        status_code, response = last_error
        return response



# -*- coding: utf-8 -*-
import json

import sysactions

from collections import OrderedDict
from bustoken import TK_FISCALWRAPPER_CANCEL_ORDER
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
    TK_SLCTRL_OMGR_VOIDORDER
)
from mw_helper import show_filtered_list_box_dialog
from posot import OrderTaker
from utilconfigurations import get_cancel_json
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union
)

from .. import mb_context
from ..util.get_menu_user_logged import get_menu_manager_authenticated_user


def void_order(
    pos_id,                 # type: Any[Union[int, str]]
    last_paid_order="0",    # type: Optional[str]
    order_id='',            # type: Optional[str]
    abandon=0,              # type: Optional[int]
    void_reason=None,       # type: Optional[Any[int, Dict]]
    pos_ot=None,            # type: Optional[OrderTaker]
    new_order_status=None,  # type: Optional[int]
):
    # type: (...) -> bool

    model = sysactions.get_model(pos_id)
    if pos_ot is None:
        pos_ot = sysactions.get_posot(model)

    order_id = int(order_id) if order_id != "" else ""
    void_order_id = order_id
    if void_reason is None:
        void_reason = get_void_reason_id(
            pos_id,
            "UNDEFINED_REASON",
            authentication_user="SYSTEM"
        )
    elif not isinstance(void_reason, dict):
        void_reason = get_fixed_void_reason(
            void_reason,
            "SYSTEM",
        )

    params = (
        '1' if pos_ot.blkopnotify else '0',
        pos_id,
        '1' if int(last_paid_order) else '0',
        void_order_id,
        '1' if abandon else '0',
        new_order_status,
        json.dumps(void_reason),
    )
    params = "\0".join(str(p) for p in params)
    msg = mb_context.MB_EasySendMessage(
        dest_name="ORDERMGR%d" % int(pos_id),
        token=TK_SLCTRL_OMGR_VOIDORDER,
        format=FM_PARAM,
        data=params,
        timeout=600000000
    )

    return msg.token == TK_SYS_ACK


def void_fiscal_order(
    pos_id,                     # type: int
    order_id,                   # type: int
    pos_id_to_display_message,  # type: int
):
    # type: (...) -> None

    data = "|".join([str(pos_id), str(order_id)])
    msg = mb_context.MB_EasySendMessage("FiscalWrapper", TK_FISCALWRAPPER_CANCEL_ORDER, format=FM_PARAM, data=data)
    if msg.token != TK_SYS_ACK:
        sysactions.show_messagebox(
            posid=pos_id_to_display_message,
            message="$ERROR_CANCEL_FISCAL_ORDER|{}".format(msg.data),
        )
        raise sysactions.StopAction(msg.data)


def get_void_reason_id(
    pos_id,                     # type: str
    selected_reason=None,       # type: Optional[str]
    void_type=None,             # type: Optional[str]
    authentication_user=None,   # type: Optional[str]
):
    # type: (...) -> Optional[Dict]

    if not selected_reason:
        void_reason_options = _get_void_reason_to_show(pos_id=pos_id, void_type=void_type)

        total_void_reasons = len(void_reason_options)
        if total_void_reasons > 1:
            title = "$SELECT_REASON"
            selected_reason = show_filtered_list_box_dialog(
                pos_id=pos_id,
                options=void_reason_options,
                title=title,
                mask="NOFILTER",
            )
        elif total_void_reasons == 1:
            selected_reason = 0

        if selected_reason is None:
            return None

        selected_reason = int(selected_reason)
        void_reason_id = void_reason_options[selected_reason].split('-')[0].strip()
        void_reason_description = void_reason_options[selected_reason].split('-')[1].strip()
        void_reason_dict = {
            'VOID_REASON_ID': void_reason_id,
            'VOID_REASON_DESCR': void_reason_description,
            "VOID_AUTH_USER": authentication_user
        }
    else:
        void_reason_dict = get_fixed_void_reason(
            reason=selected_reason,
            canceled_by=authentication_user,
        )

    return void_reason_dict


def get_fixed_void_reason(
        reason,
        canceled_by,
):
    void_reason = get_void_reason_dict().get(reason)
    void_reason_id = void_reason.get("id")
    void_reason_description = void_reason.get("reasonKey")

    void_reason_dict = {
        'VOID_REASON_ID': void_reason_id,
        'VOID_REASON_DESCR': void_reason_description,
        "VOID_AUTH_USER": canceled_by
    }

    return void_reason_dict


def void_reason_list_to_ordered_dict(void_reasons):
    # type: (...) -> OrderedDict

    void_reasons = sorted(
        void_reasons,
        key=lambda k: k["id"]
    )
    ordered_dict = OrderedDict()
    for item in void_reasons:
        ordered_dict[item["reasonKey"].replace("$", "")] = item

    return ordered_dict


def get_void_reason_dict(
    filtered_items=None,  # type: Optional[List]
):
    # type: (...) -> Dict

    void_reason_list = [
        {"id": 101, "reasonKey": "$CHANGED_YOUR_MIND"},
        {"id": 102, "reasonKey": "$DUPLICATE"},
        {"id": 103, "reasonKey": "$WRONG_SALE"},
        {"id": 104, "reasonKey": "$CANCELLATION"},
        {"id": 105, "reasonKey": "$USER_CANCELLATION"},
        {"id": 106, "reasonKey": "$USER_FISCAL_CANCELLATION"},
        {"id": 107, "reasonKey": "$TIMEOUT_CANCELLATION"},
        {"id": 108, "reasonKey": "$UNDEFINED_REASON"},
        {"id": 109, "reasonKey": "$RECALL_ORDER_TO_OLD"},
        {"id": 110, "reasonKey": "$UNKNOWN_FISCAL_ERROR"},
    ]

    void_reason_list = void_reason_list_to_ordered_dict(void_reason_list)

    return _filter_void_reason_items(void_reason_list, filtered_items)


def _filter_void_reason_items(
    void_reasons,           # type: OrderedDict
    filtered_items=None,    # type: Optional[List]
):
    # type: (...) -> Dict

    if filtered_items:
        return {k: v for k, v in void_reasons.iteritems() if v.get("id") in filtered_items}

    return void_reasons


def _get_void_reason_to_show(
    pos_id,     # type: str
    void_type,  # type: str
):
    # type: (...) -> List[str]

    void_reason_dict = get_void_reason_dict(filtered_items=[101, 102, 103, 104])
    cancel_json = get_cancel_json()
    void_reason_key = {
        "VOID_CURRENT_ORDER": "orderReasons",
        "VOID_PAID_ORDER": "couponReasons",
        "VOID_LINE": "lineReasons",
    }
    if cancel_json and void_reason_key[void_type] in cancel_json:
        void_reason_dict = void_reason_list_to_ordered_dict(cancel_json[void_reason_key[void_type]])

    return _translate_array(pos_id, void_reason_dict)


def _translate_array(
    pos_id,    # type: str
    messages,  # type: dict
):
    # type: (...) -> List[str]

    model = sysactions.get_model(pos_id)
    result = []
    for key, item in messages.items():
        reason_id = item["id"]
        reason = item["reasonKey"].encode('utf-8')
        if reason.startswith("$"):
            reason = sysactions.translate_message(
                model=model,
                message_key=reason.replace("$", ""),
            )
        result.append("{} - {}".format(
            reason_id,
            reason,
        ))

    return result


def get_canceled_by(
    pos_id,       # type: str
    model,        # type: eTree.XML
    auth_user,    # type: str
):
    # type: (...) -> str

    authenticated_manager = get_menu_manager_authenticated_user(int(pos_id))
    if authenticated_manager is not None:
        return str(authenticated_manager)

    if auth_user:
        return str(auth_user)

    operator = model.find("Operator")
    return operator.get("id", "UNDEFINED") if operator is not None else "UNDEFINED"



from util_eft import (
    get_tender_details,
    check_eft_transaction,
    get_electronic_payment_component_name,
    get_electronic_tenders,
    get_tender_id,
    has_electronic_payment,
    save_cancel_receipt,
    get_total_electronic_tenders,
)
from finish_eft_transactions import finish_eft_transactions
from request_eft import request_eft
from cancel_eft_transaction import cancel_eft_transactions
from confirm_finish_eft_transactions import confirm_finish_eft_transactions



import errno
import logging
import os
import re
import time

from datetime import datetime
from dateutil import relativedelta
from msgbus import (
    TK_SYS_NAK,
    TK_POS_GETPOSLIST,
    FM_PARAM,
    MBEasyContext
)
from typing import (
    Any,
    List,
    Optional,
    Union
)

from model.exception import PosListException

logger = logging.getLogger("Maintenance")


class CommonService(object):

    def __init__(
        self,
        mb_context,             # type: MBEasyContext
        months_to_keep_data,    # type: int
    ):
        # type: (...) -> None

        self.mb_context = mb_context
        self.months_to_keep_data = months_to_keep_data

        self.pos_list = None  # type: Optional[List[str]]

    def get_pos_list(
        self
    ):
        # type: () -> List[str]

        if self.pos_list:
            return self.pos_list

        try:
            msg = self.mb_context.MB_EasySendMessage('PosController', TK_POS_GETPOSLIST, FM_PARAM)
            if msg.token == TK_SYS_NAK or not msg.data:
                raise PosListException()

            self.pos_list = msg.data.split("\0")
            self.pos_list.sort(key=lambda x: int(x))
            return self.pos_list
        except Exception:
            logger.exception("Error getting the list of POS")

    @staticmethod
    def delete_files(
        files_to_delete     # type: List[str]
    ):
        # type: (...) -> None

        logger.info("Delete files process started. files_to_delete: {}".format(files_to_delete))
        try:
            for file_name in files_to_delete:
                dirname = os.path.dirname(file_name)
                filename = os.path.basename(file_name)
                if os.path.isdir(dirname):
                    for f in os.listdir(dirname):
                        if re.search(filename, f):
                            os.remove(os.path.join(dirname, f))
        except OSError as _:
            logger.exception("Error deleting files. files_to_delete: {}".format(files_to_delete))

    @staticmethod
    def make_sure_path_exists(
        path    # type: str
    ):
        # type: (...) -> None

        try:
            os.makedirs(path)
        except OSError as ex:
            if ex.errno != errno.EEXIST:
                raise

    def get_end_date(
        self,
        timestamp=False,    # type: Optional[bool]
    ):
        # type: (...) -> Any[Union[str, int]]

        current_month = datetime.now().replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        end_date = current_month - relativedelta.relativedelta(months=self.months_to_keep_data)

        if timestamp:
            return int(time.mktime(end_date.timetuple()))

        return end_date.strftime("%Y%m%d")

    @staticmethod
    def get_file_size(
        file_name   # type: str
    ):
        # type: (...) -> int

        return os.path.getsize(file_name) / 1024 / 1024



from typing import (
    List,
    Optional,
    Dict,
)

from ._TableOrderSaleLine import TableOrderSaleLine  # noqa
from ._OrderStateEnum import OrderStateEnum  # noqa


class TableOrder(object):

    def __init__(
        self,
        order_id,                 # type: str
        order_state,              # type: OrderStateEnum
        total_amount,             # type: float
        total_after_discount,     # type: str
        total_tender,             # type: str
        discount_amount,          # type: str
        sales_line,               # type: List[TableOrderSaleLine]
        split_key,                # type: str
        tip,                      # type: float
        discount_groups=None,     # type: Optional[Dict]
    ):
        # type: (...) -> None

        self.order_id = order_id
        self.order_state = order_state
        self.total_amount = total_amount
        self.total_after_discount = total_after_discount
        self.total_tender = total_tender
        self.discount_amount = discount_amount
        self.sales_line = sales_line
        self.split_key = split_key
        self.tip = tip
        self.discount_groups = discount_groups



from __future__ import absolute_import

# For backwards compatibility, provide imports that used to be here.
from .connection import is_connection_dropped
from .request import make_headers
from .response import is_fp_closed
from .ssl_ import (
    SSLContext,
    HAS_SNI,
    IS_PYOPENSSL,
    IS_SECURETRANSPORT,
    assert_fingerprint,
    resolve_cert_reqs,
    resolve_ssl_version,
    ssl_wrap_socket,
    PROTOCOL_TLS,
)
from .timeout import current_time, Timeout

from .retry import Retry
from .url import get_host, parse_url, split_first, Url
from .wait import wait_for_read, wait_for_write

__all__ = (
    "HAS_SNI",
    "IS_PYOPENSSL",
    "IS_SECURETRANSPORT",
    "SSLContext",
    "PROTOCOL_TLS",
    "Retry",
    "Timeout",
    "Url",
    "assert_fingerprint",
    "current_time",
    "is_connection_dropped",
    "is_fp_closed",
    "get_host",
    "parse_url",
    "make_headers",
    "resolve_cert_reqs",
    "resolve_ssl_version",
    "split_first",
    "ssl_wrap_socket",
    "wait_for_read",
    "wait_for_write",
)



import base64

from flask import Response, Request
from messagebus import Message
from typing import Dict


class SendMessageProcessor(object):
    def __init__(self, message_bus, min_deflate_size, logger):
        self.message_bus = message_bus
        self.logger = logger
        self.min_deflate_size = min_deflate_size

    def handle(self, path_parameters, request):
        # type: (Dict[str, str], Request) -> Response
        service_name = path_parameters['serviceName']
        token_request = request.args['token']
        if token_request == "TK_POS_EXECUTEACTION":
            token_request = int("0xF0500003", 16)
        else:
            token_request = int(token_request, 16)

        format_request = int(request.args['format'])
        if 'timeout' in request.args:
            timeout_request = int(request.args['timeout']) * 1000
        else:
            timeout_request = -1
        is_base_64_request = request.args['isBase64']
        if 'payload' in request.args:
            body_request = request.args['payload']
        else:
            body_request = request.get_data(parse_form_data=False)

        try:
            if is_base_64_request == 'true' and len(body_request) > 0:
                body_request = base64.b64decode(body_request)

            message = Message(int(token_request), int(format_request), body_request, int(timeout_request))
            reply = self.message_bus.send_message(service_name, message)

            if len(reply.data) < self.min_deflate_size:
                return Response(reply.data, status=200, content_type="text")
            else:
                return Response(reply.data, status=200, content_type="application/xml")
        except Exception as err:
            if self.logger:
                self.logger.exception("Request from {} - Error sending Message".format(request.remote_addr))

            return Response(err, status=400)



import json

from xml.etree import cElementTree as eTree

from sysactions import get_model, get_custom

from application.model.order import Order


def get_service_period(order):
    # type: (Order) -> (str, eTree.XML)

    model = get_model(order.pos_id)
    service_period = get_custom(model, 'SERVICE_PERIOD') if model else None
    service_period = json.loads(service_period) if service_period else None

    if order.table_id:
        order.get_table_info()
        if order.special_catalog:
            return order.special_catalog
        if service_period and order.sector in service_period['sectors'].keys():
            return service_period['sectors'][order.sector]

    if service_period:
        return service_period

    return {}



# -*- coding: utf-8 -*-
import json

import sysactions
from actions.util import get_ruptured_products_from_remote_order
from bustoken import TK_REMOTE_ORDER_GET_RUPTURED_ITEMS
from msgbus import TK_SYS_NAK, FM_PARAM
from typing import List

from .. import logger
from .. import mb_context


@sysactions.action
def get_ruptured_products(pos_id):
    # type: (str) -> str

    try:
        products = {
                "enabledProducts": [],
                "disabledProducts": get_ruptured_products_from_remote_order()
        }

        return json.dumps(products)

    except Exception as e:
        logger.exception('[get_ruptured_products] Exception {}'.format(e))
        sysactions.show_messagebox(pos_id, "$RUPTURE_ERROR", "$INFORMATION")


def _get_ruptured_products_from_json():
    # type: () -> List

    products_list = []
    ruptured_products = _get_ruptured_products(products_list)
    return ruptured_products


def _get_ruptured_products(products):
    # type: (List) -> List

    token = TK_REMOTE_ORDER_GET_RUPTURED_ITEMS
    msg = mb_context.MB_EasySendMessage("RemoteOrder", token, FM_PARAM, json.dumps(products))
    if msg.token == TK_SYS_NAK:
        raise Exception(msg.data)

    return json.loads(msg.data)


def has_rupture_items(pos_id):
    # type: (int) -> None

    rupture_items = get_ruptured_products_from_remote_order()
    if not rupture_items:
        return

    rupture_reset_alert = '$RUPTURE_RESET_UPDATE'
    sysactions.show_messagebox(pos_id, message=rupture_reset_alert, buttons="$OK")



from abc import ABCMeta, abstractmethod
from typing import (
    Optional,
    List,
)
from xml.etree import cElementTree as eTree

from posot import OrderTaker

from interactor.route.payment.insert.dto import OrderInsertPaymentRequest
from interactor.route.electronic_payment.send_responses.dto import SendResponsesRequest


class FiscalServiceModel(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def process_order_fiscal(
        self,
        model,              # type: eTree
        pos_ot,             # type: OrderTaker
        print_coupon=True,  # type: bool
    ):
        # type: (...) -> Optional[str]

        raise NotImplementedError

    @abstractmethod
    def process_payment(self, model, request):
        # type: (eTree, OrderInsertPaymentRequest) -> None

        raise NotImplementedError

    def cancel_electronic_payment(
        self,
        model,      # type: eTree
        request,    # type: OrderInsertPaymentRequest
    ):
        # type: (...) -> None

        raise NotImplementedError

    @abstractmethod
    def send_electronic_payment_message_response(self, model, request):
        # type: (eTree, SendResponsesRequest) -> None

        raise NotImplementedError

    def get_customer_receipts(
        self,
        model,      # type: eTree
    ):
        # type: (...) -> List[str]

        raise NotImplementedError




from ._CashReportFilterDto import CashReportFilterDto

from typing import Optional  # noqa
from datetime import datetime  # noqa


class PeriodCashReportFilterDto(CashReportFilterDto):
    def __init__(self, pos_id, report_type, initial_date, end_date, pos=None, operator_id=None):
        # type: (int, str, datetime, datetime, Optional[int], Optional[int]) -> None
        super(PeriodCashReportFilterDto, self).__init__(report_type)
        self.pos_id = pos_id
        self.initial_date = initial_date
        self.end_date = end_date
        self.pos = pos
        self.operator_id = operator_id



u"""
Fixer for "class Foo: ..." -> "class Foo(object): ..."
"""

from lib2to3 import fixer_base
from lib2to3.fixer_util import LParen, RParen, Name

from libfuturize.fixer_util import touch_import_top


def insert_object(node, idx):
    node.insert_child(idx, RParen())
    node.insert_child(idx, Name(u"object"))
    node.insert_child(idx, LParen())

class FixNewstyle(fixer_base.BaseFix):

    # Match:
    #   class Blah:
    # and:
    #   class Blah():

    PATTERN = u"classdef< 'class' NAME ['(' ')'] colon=':' any >"

    def transform(self, node, results):
        colon = results[u"colon"]
        idx = node.children.index(colon)
        if (node.children[idx-2].value == '(' and
            node.children[idx-1].value == ')'):
            del node.children[idx-2:idx]
            idx -= 2
        insert_object(node, idx)
        touch_import_top(u'builtins', 'object', node)



from abc import ABCMeta
from typing import Dict

from _AposA8Align import AposA8Align
from _AposA8CommandType import AposA8CommandType
from _AposA8PrintLine import AposA8PrintLine
from _AposA8Helpers import (
    append_items_header,
    append_and_count_sale_items,
    APOS_DASH,
    append_coupon_fiscal_parameters,
    append_customer_name,
)
from _AposA8Constants import AposA8Constants
from printer import FiscalPrinter
from msgbus import MBEasyContext

from clichehelper import ClicheHelper

AMOUNT_TEMPLATE = "%-36s   R$%7.2f"


class AposA8FiscalPrinter(FiscalPrinter):
    __metaclass__ = ABCMeta

    def __init__(
        self,
        mb_context,                     # type: MBEasyContext
        print_customer_name,            # type: bool
        cliche_helper,                  # type: ClicheHelper
        customer_text_field,            # type: str
        coupon_fiscal_parameters,       # type: Dict[str, bool]
        font_size_withdrawal_panel,     # type: str
    ):
        # type: (...) -> None

        super(AposA8FiscalPrinter, self).__init__(
            mb_context,
            print_customer_name,
            cliche_helper,
            coupon_fiscal_parameters,
        )

        self.customer_text_field = customer_text_field
        self.font_size_withdrawal_panel = font_size_withdrawal_panel

    def print_cancel(self, posid, data):
        # not implemented
        pass

    def _append_total(self, report, order, tenders, items_amount):
        qty_total = "%-36s         %3d" % ("QTD. TOTAL DE ITENS", items_amount)
        report.append(AposA8PrintLine(AposA8Align.CENTER, qty_total, AposA8CommandType.PRINT_TEXT))

        total_amount_value = float(order.get("totalAmount")) + float(order.get("taxTotal")) + float(order.get("tip", 0))

        total_amount_text = AMOUNT_TEMPLATE % ("VALOR TOTAL R$", total_amount_value)
        report.append(AposA8PrintLine(AposA8Align.CENTER, total_amount_text, AposA8CommandType.PRINT_TEXT))

        discount_amount_value = float(order.get("discountAmount"))
        discount = AMOUNT_TEMPLATE % ("DESCONTO", discount_amount_value)
        report.append(AposA8PrintLine(AposA8Align.CENTER, discount, AposA8CommandType.PRINT_TEXT))

        total_payable = AMOUNT_TEMPLATE % ("VALOR A PAGAR", total_amount_value - discount_amount_value)
        report.append(AposA8PrintLine(AposA8Align.CENTER, total_payable, AposA8CommandType.PRINT_TEXT))

        payment = "%-36s  %s" % ("FORMA DE PAGAMENTO", "Valor Pago")
        report.append(AposA8PrintLine(AposA8Align.CENTER, payment, AposA8CommandType.PRINT_TEXT))
        for tender in tenders:
            tender_descr = self.get_payment_description(self.mb_context, tender)
            tender_amount = tender.amount + tender.change
            tender_text = AMOUNT_TEMPLATE % (tender_descr, tender_amount)
            report.append(AposA8PrintLine(AposA8Align.CENTER, tender_text, AposA8CommandType.PRINT_TEXT))

        # Get "Change" (sixth value of tender tuple) from last tender in tenders
        change = tenders[len(tenders) - 1].change
        if change > 0:
            order_voided = order.get("state") == "VOIDED"
            factor = -1 if order_voided else 1

            change_text = AMOUNT_TEMPLATE % ("Troco", change * factor)
            report.append(AposA8PrintLine(AposA8Align.CENTER, change_text, AposA8CommandType.PRINT_TEXT))

    def _append_customer_text_field(self, report):
        if self.customer_text_field != "":
            for customer_text in self.customer_text_field.split('\\n'):
                report.append(
                    AposA8PrintLine(AposA8Align.CENTER, customer_text.replace("\\n", ""), AposA8CommandType.PRINT_TEXT))
            report.append(
                AposA8PrintLine(AposA8Align.CENTER, AposA8Constants.SEPARATOR_DASH, AposA8CommandType.PRINT_TEXT))

    @staticmethod
    def _append_satisfaction_survey(report, model, satisfaction_survey_code):
        if satisfaction_survey_code is not None and model is not None:
            report.append(APOS_DASH)
            ssc = satisfaction_survey_code
            satisfaction_survey_code = '-'.join(ssc[i:i + 3] for i in range(0, len(ssc), 3))
            report.append(AposA8PrintLine(AposA8Align.CENTER, satisfaction_survey_code, AposA8CommandType.PRINT_TEXT))
            report.append(APOS_DASH)

    def _append_coupon_body(self, customer_name, model, order, report, tenders, printer_common):
        append_items_header(report)
        items_amount = append_and_count_sale_items(report, order, printer_common)
        report.append(APOS_DASH)
        self._append_total(report, order, tenders, items_amount)
        report.append(APOS_DASH)

        append_customer_name(
            model=model,
            print_customer_name=self.print_customer_name,
            report=report,
            order=order,
            customer_name=customer_name,
            font_size_withdrawal_panel=self.font_size_withdrawal_panel,
        )

        append_coupon_fiscal_parameters(
            coupon_fiscal_parameters=self.coupon_fiscal_parameters,
            report=report,
            order=order,
        )

        self._append_satisfaction_survey(report, model, self.get_satisfaction_survey_code(order))

    def _append_cliche(self, report):
        for line in self.cliche_helper.get_cliche():
            report.append(AposA8PrintLine(AposA8Align.CENTER, line, AposA8CommandType.PRINT_TEXT))
        report.append(APOS_DASH)






from abc import ABCMeta, abstractmethod
from typing import List, Dict

from model.payment import PaymentType


class PaymentServiceModel(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_payment_types(
            self,
    ):
        # type: (...) -> List[PaymentType]

        raise NotImplementedError

    @abstractmethod
    def format_payment_types(
        self,
        tender_types    # type: List[Dict[int, str]]
    ):
        # type: (...) -> List[PaymentType]

        raise NotImplementedError

    def get_tender_details(
        self,
        payment_data,    # type: OrderInsertPaymentRequest
    ):
        # type: (...) -> str

        raise NotImplementedError



import json
from typing import (
    Optional,
    Tuple
)

from messagebus import DefaultToken, DataType, Message
from messageprocessorutil import UuidMessageProcessor

from application.interactor import OnTableOpenInteractor
from application.model import Configurations


class OnTableOpenProcessor(UuidMessageProcessor):
    def __init__(self, configs, interactor):
        # type: (Configurations, OnTableOpenInteractor) -> None
        super(OnTableOpenProcessor, self).__init__()
        self.logger = configs.logger
        self.configs = configs
        self.interactor = interactor

    def parse_data(self, data, event_type):
        # type: (str, str) -> int

        data = json.loads(data)
        pos_id = int(data.get("pos_id"))
        table_id = int(data.get("table_id"))
        return pos_id, table_id

    def call_business(
            self,
            params  # type: Tuple[int, str]
    ):
        # type: (...) -> Optional[str]

        pos_id, table_id = params
        return self.interactor.execute(pos_id, table_id)

    def format_response(self, message_bus, message, event, request, response):
        if message:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.string, response))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        if message:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        if message:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



import os

import sysactions
from configurator import Configurator
from mbcontextmessagehandler import MbContextMessageBus, MbContextMessageHandler
from messagehandlerbuilder import ServiceMessageHandlerBuilder
from msgbus import MBEasyContext


from model.event import ListenedEvents


def main():
    # type: () -> None

    loader_path = os.environ["LOADERCFG"]

    configurator = Configurator(loader_path)
    configs = configurator.read_loader_configs()
    logger = configurator.configure_logger()

    mb_context = MBEasyContext(configs.component_name)
    message_bus = MbContextMessageBus(mb_context)
    sysactions.mbcontext = mb_context

    service_message_handler_builder = ServiceMessageHandlerBuilder(logger, configs, message_bus)
    message_handler = MbContextMessageHandler(
        message_bus,
        configs.component_name,
        configs.component_name,
        configs.required_services,
        service_message_handler_builder
    )

    listened_events = [
        ListenedEvents.ELECTRONIC_PAYMENT_SHOW_CHOICES,
        ListenedEvents.ELECTRONIC_PAYMENT_SHOW_MESSAGE,
        ListenedEvents.ELECTRONIC_PAYMENT_SHOW_QUESTION,
        ListenedEvents.ELECTRONIC_PAYMENT_SHOW_QRCODE,
        ListenedEvents.ELECTRONIC_PAYMENT_SHOW_INPUT,
        ListenedEvents.ELECTRONIC_PAYMENT_CLOSE_DIALOGS,
    ]
    message_handler.subscribe_queue_events(listened_events)

    logger.debug("Starting messagebus [{}] ...".format(configs.component_name))
    return message_handler.handle_events()


if __name__ == "__main__":
    main()



import base64
import logging
import time
from datetime import datetime
from threading import Thread
from xml.etree import cElementTree as eTree

import msgbus
import persistence
from persistence import Driver
from typing import List, Tuple, Optional, Union
from sysactions import get_model, get_posot
from old_helper import OrderTaker

from services import OrderService

logger = logging.getLogger("FiscalWrapper")


class FiscalDataRepository:
    def __init__(self, mbcontext):
        # type: (msgbus.MBEasyContext) -> FiscalDataRepository
        self.mbcontext = mbcontext
        self.conn = None  # type: Optional[persistence.Connection]
        self.order_service = OrderService()

    def __enter__(self):
        self.conn = Driver().open(self.mbcontext, service_name="FiscalPersistence")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            self.conn.close()

    def start_save_fiscal_data_thread(self, pos_id, order, xml_request, numero_sat, sent_to_nfce, xml_response):
        self.set_fiscal_xml_custom_property(pos_id, xml_request, order.get("orderId", ""))
        save_fiscal_data_thread = Thread(target=self.save_fiscal_data(pos_id, order, xml_request, numero_sat, sent_to_nfce, xml_response))
        save_fiscal_data_thread.daemon = True
        save_fiscal_data_thread.start()

    def set_fiscal_xml_custom_property(self, pos_id, xml_request, order_id='', key='FISCAL_XML'):
        # type: (str, str, str, str) -> None

        self.order_service.set_custom_property(
            pos_id=pos_id,
            key=key,
            value=xml_request,
            order_id=order_id,
            blk_notify=True,
        )

    def notify_pump(
            self,
            pos_id='',                      # type: str
            order_id='',                    # type: str
            xml_format='CancelledOrder',    # type: str
            order_picture=None              # type: Union[None, str]
    ):
        # type: (...) -> None

        try:
            event_type = "ORDER_REPROCESS"
            if not order_picture:
                order_picture = OrderTaker().get_order_picture(pos_id, order_id)

            xml = "<{}>{}</{}>".format(xml_format, order_picture, xml_format)
            self.mbcontext.MB_EasyEvtSend(event_type, event_type, xml)
        except Exception as _:
            logger.exception("Error notifying bohbump on POS ID: {} for ORDER ID: {}".format(pos_id, order_id))

    def save_fiscal_data(self, posid, order, xml_request, numero_sat, sent_to_nfce, xml_response):
        retries = 0
        while retries < 3:
            try:
                xml = base64.b64decode(xml_request)
                parsed_xml = eTree.XML(xml)

                from nfce import NfceRequestBuilder
                numero_nota_element = parsed_xml.find("infCFe/ide/nCFe")
                if numero_nota_element is None:
                    # TODO: Obter o numero da nota via parametro e remover esta logica daqui
                    numero_nota_element = parsed_xml.find(
                        "{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}nNF".format(
                            NfceRequestBuilder.NAMESPACE_NFE))
                    if numero_nota_element is None:
                        numero_nota_element = parsed_xml.find(
                            "{{{0}}}infNFe/{{{0}}}ide/{{{0}}}nNF".format(
                                NfceRequestBuilder.NAMESPACE_NFE))

                numero_nota = numero_nota_element.text

                data_nota_element = parsed_xml.find("infCFe/ide/dEmi")
                hora_nota_element = parsed_xml.find("infCFe/ide/hEmi")
                if data_nota_element is None:
                    dh_elment = parsed_xml.find(
                        "{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}dhEmi".format(
                            NfceRequestBuilder.NAMESPACE_NFE))
                    if dh_elment is None:
                        dh_elment = parsed_xml.find(
                            "{{{0}}}infNFe/{{{0}}}ide/{{{0}}}dhEmi".format(
                                NfceRequestBuilder.NAMESPACE_NFE))

                    data_nota_str = dh_elment.text

                    data_emissao_date_utc = data_nota_str[:19]
                    data_nota_date = datetime.strptime(data_emissao_date_utc, "%Y-%m-%dT%H:%M:%S")
                else:
                    data_nota_str = data_nota_element.text
                    hora_nota_str = hora_nota_element.text
                    data_nota_date = datetime.strptime(data_nota_str + hora_nota_str, "%Y%m%d%H%M%S")

                data_nota_seconds = int(time.mktime(data_nota_date.timetuple()))

                params_fiscal = {
                    "pos_id": posid,
                    "order_id": order.get("orderId"),
                    "data_nota": data_nota_seconds,
                    "numero_nota": numero_nota,
                    "numero_sat": numero_sat,
                    "xml_request": xml_request,
                    "sent_to_nfce": sent_to_nfce,
                    "xml_response": xml_response
                }

                self.conn.pquery("fiscal_saveFiscalData", **params_fiscal)
                break
            except Exception as _:
                logger.exception("Error saving fiscal data on database. RetryNum: {}".format(retries))
                time.sleep(0.2)
                retries += 1
        else:
            logger.error(">>> Save fiscal data failed after maximum retries!")

    def set_order_picture(self, order_id, order_picture):
        # type: (str, str) -> None
        params = {
            "order_id": order_id,
            "order_picture": order_picture
        }
        self.conn.pquery("fiscal_setOrderPicture", **params)

    def set_order_canceled(self, order_id):
        # type: (str) -> None
        params = {
            "order_id": order_id,
        }
        self.conn.pquery("fiscal_setOrderCanceled", **params)

    def get_nfce_orders_to_send(self, limit):
        # type: (int) -> List
        params_fiscal = {
            "limit": limit
        }
        orders = [Order(x.get_entry(0), x.get_entry(1), x.get_entry(2), x.get_entry(3), x.get_entry(4)) for
                  x in self.conn.pselect("fiscal_getNfceToSend", **params_fiscal)]
        return orders

    def get_count_nfce_orders_to_send(self):
        # type: () -> int
        return [int(x.get_entry(0)) for x in self.conn.pselect("fiscal_getNfceToSendCount")][0]

    def set_nfce_sent(self, order_id, status):
        # type: (str, int) -> None
        params_fiscal = {
            "status": status,
            "order_id": order_id
        }
        self.conn.pquery("fiscal_setNfceSent", **params_fiscal)

    def set_xmlresponse_contingency(self, order_id, xml_base64):
        # type: (str, str) -> None
        params_fiscal = {
            "order_id": order_id,
            "xml_response": xml_base64
        }
        self.conn.pquery("fiscal_setXmlResponse", **params_fiscal)

    def set_nfce_sent_with_xml(self, order_id, xml_base64, status):
        # type: (str, str, int) -> None
        params_fiscal = {
            "status": status,
            "xml_request": xml_base64,
            "order_id": order_id
        }
        self.conn.pquery("fiscal_setNfceSentXml", **params_fiscal)

    def set_nfce_sent_with_xml_bkoffice(self, order_id, xml_base64, status, bkoffice_status):
        # type: (str, str, int, int) -> None
        params_fiscal = {
            "status": status,
            "xml_request": xml_base64,
            "order_id": order_id,
            "sent_to_bkoffice": bkoffice_status
        }
        self.conn.pquery("fiscal_setNfceSentXmlBKOffice", **params_fiscal)

    def retry_fiscal_orders_with_exception(self, range_of_days):
        # type: (str) -> None
        params_fiscal = {
            "range_of_days": "-{} days".format(range_of_days),
        }
        self.conn.pquery("fiscal_retryFiscalOrdersWithException", **params_fiscal)

    def get_xml_request(self, order_id):
        # type: (str) -> str
        params_fiscal = {
            "order_id": order_id
        }
        cursor = self.conn.pselect("fiscal_getRequestXMLFiscalData", **params_fiscal)
        if cursor.rows() > 0:
            ret = [x.get_entry(0) for x in cursor][0]
            return ret

    def get_xml_response(self, order_id):
        # type: (str) -> str
        params_fiscal = {
            "order_id": order_id
        }
        cursor = self.conn.pselect("fiscal_getResponseXMLFiscalData", **params_fiscal)
        if cursor.rows() > 0:
            ret = [x.get_entry(0) for x in cursor][0]
            return ret

    def get_numero_sat_and_xml_request(self, order_id):
        # type: (str) -> Tuple
        params_fiscal = {
            "order_id": order_id
        }
        cursor = self.conn.pselect("fiscal_getSatNumberAndRequestXMLFiscalData", **params_fiscal)
        if cursor.rows() > 0:
            ret = [(x.get_entry(0), x.get_entry(1)) for x in cursor][0]
            return ret


class Order:
    def __init__(self, order_id, pos_id, xml, num_nota, order_picture):
        self.order_id = order_id
        self.pos_id = pos_id
        self.xml = xml
        self.num_nota = num_nota
        self.order_picture = order_picture


class Tender(object):
    def __init__(self):
        # type: () -> Tender
        self.tender_seq_id = ""
        self.type = ""
        self.amount = ""
        self.change = ""
        self.bandeira = ""
        self.auth_code = ""
        self.cnpj_auth = ""
        self.receipt_customer = ""
        self.receipt_merchant = ""



from _ContextKeys import ContextKeys
from _SatXmlPartBuilder import SatXmlPartBuilder
from pos_model import Order
from typing import Dict, Any


class DetListBuilder(SatXmlPartBuilder):

    def __init__(self, det_builder, det_tip_builder):
        self.det_builder = det_builder
        self.det_tip_builder = det_tip_builder

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> str
        xml = ""

        context[ContextKeys.total_vbc] = 0
        context[ContextKeys.total_prod] = 0
        context[ContextKeys.total_icms] = 0
        context[ContextKeys.total_pis] = 0
        context[ContextKeys.total_cofins] = 0
        context[ContextKeys.item_no] = 1

        for sale_item in order.sale_items:
            context[ContextKeys.current_sale_item] = sale_item
            xml += self.det_builder.build_xml(order, context)
            context[ContextKeys.item_no] += 1

        xml += self.det_tip_builder.build_xml(order, context)

        return xml



# encoding: utf-8
from typing import Dict


class TefCancellationBodyDto(object):
    def __init__(
        self,
        tenders_by_date,         # type: Dict
        card_brands     # type: Dict
    ):
        # type: (...) -> None

        self.tenders_by_date = tenders_by_date
        self.card_brands = card_brands



from application.model import Configurations
from application.service import BenefitService
from application.service import OrderService


class OnOrderTotalInteractor(object):
    def __init__(self, configs, benefit_service, order_service):
        # type: (Configurations, BenefitService, OrderService) -> None

        self.configs = configs
        self.logger = configs.logger
        self.benefit_service = benefit_service
        self.order_service = order_service

    def execute(self, pos_id):
        # type: (int) -> None

        order_picture = self.benefit_service.apply_automatic_benefits_on_order(pos_id, on_total=True)
        if not order_picture:
            order_picture = self.order_service.get_order_picture(pos_id)

        if order_picture.get("state") == "TOTALED":
            product_groups = self.benefit_service.product_groups
            order_picture = self.benefit_service.validate_benefits(pos_id, order_picture, product_groups)
            order_picture = self.benefit_service.remove_invalid_benefits(pos_id, order_picture)
            self.benefit_service.apply_pending_benefits(pos_id, order_picture)
            self.benefit_service.apply_manual_pending_benefits(pos_id, order_picture)



# encoding: utf-8

import datetime
import json
import time
from cStringIO import StringIO
from typing import (
    Dict,
    List,
    Optional,
)
from xml.etree import cElementTree as eTree

from sysactions import get_user_information

from domain import TransferType
from model import Account
from old_helper import convert_from_utf_to_localtime
from repository import (
    AccountRepository,
    ProductRepository,
)
from utils import center, join, double_line_separator, add_amount_line, empty_line, add_table_line, add_key_value_line
from utils.report_formatters import remove_accents


DATE_TIME_FMT = "%d/%m/%Y %H:%M:%S"


class CloseOperatorReport(object):
    def __init__(
        self,
        product_repository,  # type: ProductRepository
        account_repository,  # type: AccountRepository
        store_id,            # type: str
        bordereau_enabled,   # type: bool
    ):
        # type: (...) -> None
        self.product_repository = product_repository
        self.account_repository = account_repository
        self.store_id = store_id
        self.bordereau_enabled = bordereau_enabled

    def generate_close_operator_report(self, pos_id, operator_id, period, session_id, cash_report, authorizer):
        report = StringIO()
        report.write(self._close_operator_report_header(int(pos_id), operator_id, period, authorizer, session_id))
        self._close_operator_report_body(report, session_id, cash_report)

        return report.getvalue()

    def _close_operator_report_header(self, pos_id, operator_id, period, authorizer, session_id):
        title = center("Logout Operador")
        current_datetime = time.strftime(DATE_TIME_FMT)
        formatted_store_id = self.store_id
        fiscal_date = "%02d/%02d/%04d" % (int(period[6:8]), int(period[4:6]), int(period[:4]))
        separator = double_line_separator()

        user_information = get_user_information(operator_id)
        operator_name = operator_id
        if user_information is not None:
            operator_name = eTree.XML(get_user_information(operator_id)).find(".//user").get("LongName")[:15]

        authorizer_text = ""
        if authorizer is None:
            authorizer = self.account_repository.get_authorizer(session_id)
            if authorizer is not None:
                authorizer_text = "Autorizador..: {:s}".format(authorizer)
        else:
            user_information = get_user_information(authorizer)
            authorizer_name = authorizer
            if user_information is not None:
                authorizer_name = eTree.XML(get_user_information(authorizer)).find(".//user").get("LongName")[:15]
                authorizer_text = "Autorizador..: {:s}".format(authorizer_name)

        return ("""
%(separator)s
%(title)s
Loja.........: %(formatted_store_id)s
Data/hora....: %(current_datetime)s
Data fiscal..: %(fiscal_date)s
Operador.....: %(operator_name)s (POS # %(pos_id)02d)
%(authorizer_text)s
%(separator)s""" % join(globals(), locals()))

    def _close_operator_report_body(self, report, session_id, cash_report):
        tenders_description = self.product_repository.get_tenders_descripton()
        transfers = self.account_repository.get_all_transfers(session_id)
        declared_amounts = [
            transfer for transfer in transfers
            if transfer.account_type == TransferType.DECLARED_AMOUNT
        ]

        report.write(center("RESUMOS"))
        report.write(cash_report)
        self._write_cash_in_and_cash_out(
            transfers,
            declared_amounts,
            report
        )

        report.write(double_line_separator())

        if self.bordereau_enabled:
            self._write_declared_amounts(
                transfers,
                declared_amounts,
                tenders_description,
                report,
            )
        else:
            self._write_simple_summary(
                declared_amounts,
                report,
            )

    def _write_cash_in_and_cash_out(
        self,
        transfers,         # type: List[Account]
        declared_amounts,  # type: List[Account]
        report,            # type: StringIO
    ):
        # type: (...) -> None

        initial_transfer = self._get_total_transfer_by_type(transfers, TransferType.INITIAL_AMOUNT)
        total_cash_in = self._get_total_transfer_by_type(transfers, TransferType.TRANSFER_CASH_IN)
        total_cash_out = self._get_total_transfer_by_type(transfers, TransferType.TRANSFER_CASH_OUT)
        total_cash = self._get_total_transfer_by_type(
            transfers,
            TransferType.TRANSFER_SKIM,
            tender_id=0
        )

        report.write(center("SUPRIMENTOS/SANGRIAS"))
        report.write(add_amount_line("Fundo Inicial", None, initial_transfer))

        in_out_types = (TransferType.TRANSFER_CASH_IN, TransferType.TRANSFER_CASH_OUT)
        all_cash_in_and_out = [transfer for transfer in transfers if transfer.account_type in in_out_types]

        for cash in all_cash_in_and_out:
            cash_type = "Suprimento {}h"
            if cash.account_type == TransferType.TRANSFER_CASH_OUT:
                cash_type = "Sangria    {}h"

            date = datetime.datetime.strptime(cash.operation_date, "%Y-%m-%d %H:%M:%S")
            date = convert_from_utf_to_localtime(date)
            hour = str(date.hour).zfill(2) + ":" + str(date.minute).zfill(2)
            report.write(add_amount_line(cash_type.format(hour), None, cash.amount))

        drawer_amount = initial_transfer + total_cash + total_cash_in - total_cash_out
        declared_cash_amount = [transfer for transfer in declared_amounts if transfer.tender_id == 0]
        declared_cash_amount = 0.0 if len(declared_cash_amount) == 0 else declared_cash_amount[0].amount
        report.write(add_amount_line("Gaveta", None, drawer_amount))
        report.write(add_amount_line("Declarado", None, declared_cash_amount))
        report.write(add_amount_line("Diferenca", None, self._invert_sign(drawer_amount - declared_cash_amount)))

    def _write_declared_amounts(
        self,
        transfers,              # type: List[Account]
        declared_amounts,       # type: List[Account]
        tenders_description,    # type: Dict[str, str]
        report,                 # type: StringIO
    ):
        # type: (...) -> None

        report.write(center("BORDEREAU"))
        header = ["Pagamento", "Declarado", "Diferenca"]
        columns_height = [18, 10, 10]
        align = ['left', 'right', 'right']
        format_type = ['string', 'float', 'float']
        character_to_add = [".:", " ", " "]
        report.write(add_table_line(header, columns_height, align))
        total_in = 0
        total_declared = 0
        gl_account = None
        missing_tender_ids = None
        for declared in declared_amounts:
            sold_amount = self._get_total_transfer_by_type(
                transfers,
                TransferType.TRANSFER_SKIM,
                tender_id=declared.tender_id,
            )

            current_amount = sold_amount
            if declared.tender_id == 0:
                initial_transfer = self._get_total_transfer_by_type(transfers, TransferType.INITIAL_AMOUNT)
                total_cash_in = self._get_total_transfer_by_type(transfers, TransferType.TRANSFER_CASH_IN)
                total_cash_out = self._get_total_transfer_by_type(transfers, TransferType.TRANSFER_CASH_OUT)

                current_amount += initial_transfer + total_cash_in - total_cash_out
                gl_account = declared.glaccount

            if declared.is_undeclared:
                difference = 0
            else:
                total_in += current_amount
                total_declared += declared.amount
                difference = self._invert_sign(current_amount - declared.amount)

            tender_id = str(declared.tender_id)
            declared_amount = current_amount if declared.is_undeclared else declared.amount
            if tender_id in tenders_description:
                formatted_tender = remove_accents(
                    tenders_description[tender_id]
                )
                line = [formatted_tender, declared_amount, difference]
                report.write(add_table_line(line, columns_height, align, format_type, character_to_add))
            else:
                if missing_tender_ids:
                    missing_tender_ids['declared_amount'] += declared_amount
                    missing_tender_ids['difference'] += difference
                else:
                    missing_tender_ids = {
                        'declared_amount': declared_amount,
                        'difference': difference,
                    }

        if missing_tender_ids:
            line = [
                'Outros',
                missing_tender_ids['declared_amount'],
                missing_tender_ids['difference']
            ]
            report.write(
                add_table_line(
                    line,
                    columns_height,
                    align,
                    format_type,
                    character_to_add
                )
            )

        report.write(empty_line())
        gl_account = {} if not gl_account else json.loads(gl_account)
        if gl_account:
            diff_total = self._invert_sign(total_in - total_declared)
            report.write(add_amount_line("Diferenca total", None, diff_total))
            report.write(add_key_value_line("Numero Envelope..: ", gl_account["envelope"]))
            justification = remove_accents(gl_account["justification"])
            report.write(add_key_value_line("Justificativa....: ", justification))

    @staticmethod
    def _write_simple_summary(
        declared_amounts,   # type: List[Account]
        report,             # type: StringIO
    ):
        # type: (...) -> None

        gl_account = {}
        for declared in declared_amounts:
            if declared.tender_id == 0:
                gl_account = json.loads(declared.glaccount or "{}")
                break

        report.write(empty_line())
        report.write(add_key_value_line("Numero Envelope..: ", gl_account.get("envelope", "")))
        report.write(add_key_value_line("Justificativa....: ", gl_account.get("justification", "")))

    @staticmethod
    def _get_total_transfer_by_type(
        transfers,       # type: List[Account]
        account_type,    # type: TransferType
        tender_id=None,  # type: Optional[int]
    ):
        # type: (...) -> float

        if tender_id is None:
            total = sum(
                transfer.amount for transfer in transfers
                if transfer.account_type == account_type
            )
        else:
            total = sum(
                transfer.amount for transfer in transfers
                if transfer.account_type == account_type and transfer.tender_id == tender_id
            )
        return total

    @staticmethod
    def _invert_sign(value):
        return -value if value != 0 else value



from application.interactor import (
    BenefitRetryInteractor,
    BurnBenefitInteractor,
    GetBenefitInfoInteractor,
    ApplyBenefitInteractor,
    UnlockBenefitInteractor,
    OrderModifiedPaidInteractor,
    OrderModifiedVoidInteractor,
    OrderModifiedVoidLineInteractor,
    TestConnectionInteractor
)
from application.manager import BenefitRetryManager
from application.model import ListenedTokens, ListenedEvents, DefaultBenefitApplierRepository, BenefitAppliers
from application.model.configuration import Configurations
from application.processor import (
    GetBenefitInfoProcessor,
    ApplyBenefitProcessor,
    OrderModifiedPaidProcessor,
    OrderModifiedVoidProcessor,
    OrderModifiedVoidLineProcessor,
    NotMappedProcessor,
    TestConnectionProcessor
)
from application.repository.pos import DBRepository
from messagebus import Message, Event
from messagehandler import MessageHandlerBuilder
from messageprocessor import \
    MessageProcessorMessageHandler, \
    DefaultMessageProcessorExecutorFactory
from messageprocessorutil import LoggingProcessorCallback
from typing import Union, Dict


class ComponentMessageHandlerBuilder(MessageHandlerBuilder):

    def __init__(self, configs, db_repository, benefit_appliers_repositories):
        # type: (Configurations, DBRepository, Dict[BenefitAppliers: DefaultBenefitApplierRepository]) -> None

        self.logger = configs.logger
        self.configs = configs
        self.db_repository = db_repository
        self.benefit_appliers_repositories = benefit_appliers_repositories

        self.benefit_retry_manager = None  # type: Union[None, BenefitRetryManager]
        self.message_handler = None  # type: Union[None, MessageProcessorMessageHandler]

    def build_singletons(self):
        # type: () -> None

        default_params = [self.configs, self.db_repository, self.benefit_appliers_repositories]

        burn_benefit_interactor = BurnBenefitInteractor(*default_params)
        get_benefit_info_interactor = GetBenefitInfoInteractor(*default_params)
        apply_benefit_interactor = ApplyBenefitInteractor(*default_params)
        unlock_benefit_interactor = UnlockBenefitInteractor(*default_params)
        test_connection_interactor = TestConnectionInteractor(self.configs)

        order_modified_paid_interactor = OrderModifiedPaidInteractor(burn_benefit_interactor, *default_params)
        order_modified_void_interactor = OrderModifiedVoidInteractor(unlock_benefit_interactor, *default_params)
        order_modified_void_line_interactor = OrderModifiedVoidLineInteractor(unlock_benefit_interactor, *default_params)

        get_benefit_info_processor = GetBenefitInfoProcessor(get_benefit_info_interactor)
        apply_benefit_processor = ApplyBenefitProcessor(apply_benefit_interactor)

        order_modified_paid_processor = OrderModifiedPaidProcessor(order_modified_paid_interactor)
        order_modified_void_processor = OrderModifiedVoidProcessor(order_modified_void_interactor)
        order_modified_void_line_processor = OrderModifiedVoidLineProcessor(order_modified_void_line_interactor)
        not_mapped_processor = NotMappedProcessor()

        test_connection_processor = TestConnectionProcessor(test_connection_interactor)

        messages_processors = {
            ListenedTokens.TK_NCOUPON_APPLY_COUPON.value: apply_benefit_processor,
            ListenedTokens.TK_NCOUPON_GET_COUPON_INFO.value: get_benefit_info_processor,
            ListenedTokens.TK_NCOUPON_TEST_CONNECTION.value: test_connection_processor
        }

        events_processors = {
            ListenedEvents.ORDER_MODIFIED_PAID.value: order_modified_paid_processor,
            ListenedEvents.ORDER_MODIFIED_VOID.value: order_modified_void_processor,
            ListenedEvents.ORDER_MODIFIED_VOID_LINE.value: order_modified_void_line_processor,
            ListenedEvents.ORDER_MODIFIED.value: not_mapped_processor
        }

        callbacks = [LoggingProcessorCallback(self.logger)]

        self.message_handler = MessageProcessorMessageHandler(
            event_processors=events_processors,
            message_processors=messages_processors,
            executor_factory=DefaultMessageProcessorExecutorFactory(callbacks),
        )

        self.benefit_retry_manager = self._build_benefits_retry_manager(
            burn_benefit_interactor=burn_benefit_interactor,
            unlock_benefit_interactor=unlock_benefit_interactor,
        )
        self.benefit_retry_manager.start()

    def build_message_handler(self, message, event):
        # type: (Message, Event) -> None

        return self.message_handler

    def destroy_singletons(self):
        # type: () -> None

        self.benefit_retry_manager.stop()

    def destroy_message_handler(self, message_handler):
        # type: (MessageHandlerBuilder) -> None

        return

    def _build_benefits_retry_manager(self, burn_benefit_interactor, unlock_benefit_interactor):
        # type: (BurnBenefitInteractor, UnlockBenefitInteractor) -> BenefitRetryManager

        interactor = BenefitRetryInteractor(
            self.configs,
            self.db_repository,
            burn_benefit_interactor,
            unlock_benefit_interactor
        )
        return BenefitRetryManager(self.configs, interactor)
